// Camera.cpp: Implementierung der Klasse Camera.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Camera.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Camera::Camera()
{

	SETPARAMETERFILE("CameraParams");

	m_fClipNear = LOADFLOAT("m_fClipNear", 5.0f);
	m_fClipFar = LOADFLOAT("m_fClipFar", 50000.0f);

	m_fAspectWidth = LOADFLOAT("m_fAspectWidth", 1.0f);
	m_fAspectHeight = LOADFLOAT("m_fAspectHeight", 0.8f);

	m_vPos.x = LOADFLOAT("m_fStartX", 4000.0f);
	m_vPos.y = LOADFLOAT("m_fStartY", 5000.0f);
	m_vPos.z = LOADFLOAT("m_fStartZ", 1500.0f);

	m_fAngleX = LOADFLOAT("m_fAngleX", 0.0f);
	m_fAngleY = LOADFLOAT("m_fAngleY", 0.2f);

	m_fMoveSpeed = LOADFLOAT("m_fMoveSpeed", 8.5f);
	m_fCameraRadius = LOADFLOAT("m_fCameraRadius", 100.0f);


	ZeroMemory(&m_FrustrumPlane, sizeof(m_FrustrumPlane));

	D3DXMatrixPerspectiveLH( &m_matProjection, m_fAspectWidth*m_fClipNear,
											   m_fAspectHeight*m_fClipNear, 
											   m_fClipNear,
											   m_fClipFar );

	SetProjectionTransform();
	SetTransform();
}

void Camera::Pitch(float _alpha)
{
    m_fAngleX += _alpha;
    if (m_fAngleX >= D3DX_PI/2 || m_fAngleX <= -D3DX_PI/2) m_fAngleX -= _alpha;
}

void Camera::Yaw(float _alpha)
{
    m_fAngleY += _alpha;    
}

void Camera::Move(int _axis, float _dist)
{
    _dist *= m_fMoveSpeed;
    switch (_axis)
    {

        case 0:
            m_vPos += _dist*D3DXVECTOR3(m_matView._11, m_matView._21, m_matView._31);
            break;
        case 1:
            m_vPos += _dist*D3DXVECTOR3(m_matView._12, m_matView._22, m_matView._32);
            break;
        case 2:
            m_vPos += _dist*D3DXVECTOR3(m_matView._13, m_matView._23, m_matView._33);
            break;
    }
}

void Camera::SetTransform()
{
	float sinY = sinf(m_fAngleY);
	float sinX = sinf(m_fAngleX);
	float cosY = cosf(m_fAngleY);
	float cosX = cosf(m_fAngleX);

	D3DXMatrixIdentity(&m_matView);

	// x-axis : right
	m_matView._11 = cosY;
	m_matView._21 = 0.0f;
	m_matView._31 = -sinY;

	// y-axis : up
	m_matView._12 = -sinX*sinY;
	m_matView._22 = cosX;
	m_matView._32 = -sinX*cosY;

	// z-axis : forward
	m_matView._13 = sinY*cosX;
	m_matView._23 = sinX;
	m_matView._33 = cosY*cosX;

	// translation
	m_matView._41 = -m_matView._11*m_vPos.x - m_matView._21*m_vPos.y - m_matView._31*m_vPos.z;
	m_matView._42 = -m_matView._12*m_vPos.x - m_matView._22*m_vPos.y - m_matView._32*m_vPos.z;
	m_matView._43 = -m_matView._13*m_vPos.x - m_matView._23*m_vPos.y - m_matView._33*m_vPos.z;

	SetFrustrumPlanes();

    g_pD3DRenderLayer->SetTransform( D3DTS_VIEW, &m_matView );

}

void Camera::SetProjectionTransform()// only necessary if there are multiple cameras with different projection matrices
{
    g_pD3DRenderLayer->SetTransform( D3DTS_PROJECTION, &m_matProjection );
}

int Camera::FrustrumIntersect(AABB *_bb)
{
	D3DXVECTOR3 bMax, bMin;
	unsigned char clipcode;
	bool intersect = false;

	for (int i=0; i<6; i++)
	{

		// choose right diagonal to check
		if (m_FrustrumPlane[i].a >= 0)
		{
			bMin.x = _bb->minX;
			bMax.x = _bb->maxX;
		}
		else
		{
			bMax.x = _bb->minX;
			bMin.x = _bb->maxX;
		}

		if (m_FrustrumPlane[i].b >= 0)
		{
			bMin.y = _bb->minY;
			bMax.y = _bb->maxY;
		}
		else
		{
			bMax.y = _bb->minY;
			bMin.y = _bb->maxY;
		}

		if (m_FrustrumPlane[i].c >= 0)
		{
			bMin.z = _bb->minZ;
			bMax.z = _bb->maxZ;
		}
		else
		{
			bMax.z = _bb->minZ;
			bMin.z = _bb->maxZ;
		}

		clipcode = (m_FrustrumPlane[i].a*bMin.x + 
				    m_FrustrumPlane[i].b*bMin.y +
				    m_FrustrumPlane[i].c*bMin.z + 
				    m_FrustrumPlane[i].d > 0);
		clipcode |= (m_FrustrumPlane[i].a*bMax.x + 
				     m_FrustrumPlane[i].b*bMax.y +
					 m_FrustrumPlane[i].c*bMax.z + 
					 m_FrustrumPlane[i].d > 0)<< 1;

		if (clipcode != 0 && clipcode != 3) intersect = true;
		else
		if (clipcode == 3) return FI_OUTSIDE;
	}

	return intersect ? FI_INTERSECT : FI_INSIDE;
}

void Camera::SetFrustrumPlanes()
{
	// untransformed normals of view frustrum
	static const D3DXVECTOR3 normal[]=
	{ 
		D3DXVECTOR3( 0.0f, 0.0f,-1.0f),					    //near
		D3DXVECTOR3( 0.0f, 0.0f, 1.0f),					    //far
		D3DXVECTOR3(-1.0f, 0.0f,-m_fAspectWidth/2),  //left
		D3DXVECTOR3( 1.0f, 0.0f,-m_fAspectWidth/2),  //right
		D3DXVECTOR3( 0.0f, 1.0f,-m_fAspectHeight/2), //top
		D3DXVECTOR3( 0.0f,-1.0f,-m_fAspectHeight/2)  //bottom
	};

	D3DXMATRIX transposeView;
	D3DXMatrixTranspose(&transposeView, &m_matView);

	// use transformed vectors to get surface equation for clipping planes
	for (int i=0; i<6; i++)
	{
		D3DXVec3TransformNormal((D3DXVECTOR3*)&(m_FrustrumPlane[i]), &normal[i], &transposeView);
	}

	// get one point inside near clipping plane
	D3DXVECTOR3 tmp = m_vPos;
	tmp.x = m_matView._13 * m_fClipNear + m_vPos.x;
	tmp.y = m_matView._23 * m_fClipNear + m_vPos.y;
	tmp.z = m_matView._33 * m_fClipNear + m_vPos.z;
	m_FrustrumPlane[0].d = -(m_FrustrumPlane[0].a*tmp.x + m_FrustrumPlane[0].b*tmp.y + m_FrustrumPlane[0].c*tmp.z);

	// get one point inside far clipping plane
	tmp = m_vPos;
	tmp.x = m_matView._13 * m_fClipFar + m_vPos.x;
	tmp.y = m_matView._23 * m_fClipFar + m_vPos.y;
	tmp.z = m_matView._33 * m_fClipFar + m_vPos.z;
	m_FrustrumPlane[1].d = -(m_FrustrumPlane[1].a*tmp.x + m_FrustrumPlane[1].b*tmp.y + m_FrustrumPlane[1].c*tmp.z);

	m_FrustrumPlane[2].d = -(m_FrustrumPlane[2].a*m_vPos.x + m_FrustrumPlane[2].b*m_vPos.y + m_FrustrumPlane[2].c*m_vPos.z);
	m_FrustrumPlane[3].d = -(m_FrustrumPlane[3].a*m_vPos.x + m_FrustrumPlane[3].b*m_vPos.y + m_FrustrumPlane[3].c*m_vPos.z);
	m_FrustrumPlane[4].d = -(m_FrustrumPlane[4].a*m_vPos.x + m_FrustrumPlane[4].b*m_vPos.y + m_FrustrumPlane[4].c*m_vPos.z);
	m_FrustrumPlane[5].d = -(m_FrustrumPlane[5].a*m_vPos.x + m_FrustrumPlane[5].b*m_vPos.y + m_FrustrumPlane[5].c*m_vPos.z);
	
}

void Camera::SetTransformLookAt(D3DXVECTOR3& _at)
{
	static const D3DXVECTOR3 up = D3DXVECTOR3(0.0f, 1.0f, 0.0f);

	D3DXMatrixLookAtLH(&m_matView, &m_vPos, &_at, &up);

	SetFrustrumPlanes();

    g_pD3DRenderLayer->SetTransform( D3DTS_VIEW, &m_matView );
}

void Camera::SetPos(D3DXVECTOR3 &_pos)
{
	m_vPos = _pos;
}

D3DXVECTOR3 Camera::GetPos()
{
	return m_vPos;
}
