// D3DLayer.cpp: Implementierung der Klasse D3DLayer.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "D3DLayer.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

D3DLayer::D3DLayer(HWND hWnd, HINSTANCE hInst)
{
    g_pUtils->Log("\nD3D Initialization");

	SETPARAMETERFILE("D3DLayerParams");
	m_iResX = LOADINT("m_iResX", 1280);
	m_iResY = LOADINT("m_iResY", 1024);
	m_iBPP  = LOADINT("m_iBPP" , 32);

    HRESULT hres;

    // Create the D3D object
    if( NULL == ( m_pD3D = Direct3DCreate8( D3D_SDK_VERSION ) ) )
        throw InitException("Direct3DCreate8 FAILED!");

	
    D3DDISPLAYMODE d3ddmDisplay;
    D3DDISPLAYMODE d3ddmBackBuffer;
	char msg[64];

    if (FULLSCREEN)
    {
        // enumerate adapter modes

        int iNumModes = m_pD3D->GetAdapterModeCount(D3DADAPTER_DEFAULT);
        D3DDISPLAYMODE d3ddm;
		ZeroMemory(&d3ddmDisplay, sizeof(d3ddmDisplay));
        bool bFoundMatch = false;
        for (int i=0; i<iNumModes; i++)
        {
            m_pD3D->EnumAdapterModes(D3DADAPTER_DEFAULT, i, &d3ddm);
            if (d3ddm.Width == (unsigned)m_iResX && d3ddm.Height == (unsigned)m_iResY)
            {
                //check color format
                switch (m_iBPP)
                {
                    case 16:
                        switch (d3ddm.Format)
                        {
                            case D3DFMT_R5G6B5:
                            case D3DFMT_X1R5G5B5:

								if (d3ddm.Format == D3DFMT_R5G6B5)								
									sprintf(msg, "Enumerated mode: %dx%dxD3DFMT_R5G6B5, %d Hz", d3ddm.Width, d3ddm.Height,
										d3ddm.RefreshRate);
								else
									sprintf(msg, "Enumerated mode: %dx%dxD3DFMT_X1R5G5B5, %d Hz", d3ddm.Width, d3ddm.Height,
										d3ddm.RefreshRate);
								g_pUtils->Log(msg);


								if (d3ddm.RefreshRate >= d3ddmDisplay.RefreshRate)
								{
									d3ddmDisplay = d3ddm;
									bFoundMatch = true;
								}
								break;
                        }
                        break;
                    case 32:
                        switch (d3ddm.Format)
                        {
                            case D3DFMT_X8R8G8B8:
								sprintf(msg, "Enumerated mode: %dx%dx%d, %d Hz", d3ddm.Width, d3ddm.Height,
									32, d3ddm.RefreshRate);
								g_pUtils->Log(msg);
								if (d3ddm.RefreshRate >= d3ddmDisplay.RefreshRate)
								{
									d3ddmDisplay = d3ddm;
									bFoundMatch = true;
								}
								break;
                        }
                        break;
                    default:
                        throw InitException("Unsupported color depth specified");
                        break;
                }
            }
        }

        if (!bFoundMatch) throw InitException("No matching displaymode found!");

		sprintf(msg, "Selected display mode: %dx%dx%d", m_iResX, m_iResY, m_iBPP/*, d3ddmDisplay.RefreshRate*/);
		g_pUtils->Log(msg);
        d3ddmBackBuffer = d3ddmDisplay;

    } else //windowed mode
    {
        hres = m_pD3D->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &d3ddmDisplay );
        if (FAILED(hres)) throw InitException(hres, "GetAdapterDisplayMode FAILED!");
        d3ddmBackBuffer = d3ddmDisplay;
    }


	
    D3DPRESENT_PARAMETERS d3dpp; 
    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    
    d3dpp.BackBufferFormat = d3ddmBackBuffer.Format;

    // check hardware support
    if (FAILED(m_pD3D->CheckDeviceType(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, d3ddmDisplay.Format, d3dpp.BackBufferFormat, false)))
        throw InitException("No Hardware support!");       



    //check depth- & stencilbuffer	
	m_bStencil = true;
    if (SUCCEEDED(m_pD3D->CheckDepthStencilMatch(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, d3ddmDisplay.Format,
											     d3ddmDisplay.Format, D3DFMT_D24S8)))
    {
        d3dpp.AutoDepthStencilFormat = D3DFMT_D24S8;
		g_pUtils->Log("Depth buffer depth: 24 bit; Stencil buffer depth: 8 bit");
    } else

    if (SUCCEEDED(m_pD3D->CheckDepthStencilMatch(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, d3ddmDisplay.Format,
											     d3ddmDisplay.Format, D3DFMT_D24X4S4)))
    {
        d3dpp.AutoDepthStencilFormat = D3DFMT_D24X4S4;
		g_pUtils->Log("Depth buffer depth: 24 bit; Stencil buffer depth: 4 bit");
    } else

    if (SUCCEEDED(m_pD3D->CheckDepthStencilMatch(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, d3ddmDisplay.Format,
											     d3ddmDisplay.Format, D3DFMT_D15S1)))
    {
        d3dpp.AutoDepthStencilFormat = D3DFMT_D15S1;
		g_pUtils->Log("Depth buffer depth: 16 bit; Stencil buffer depth: 1 bit");
    } else 
	{
		m_bStencil = false;

		if (SUCCEEDED(m_pD3D->CheckDepthStencilMatch(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, d3ddmDisplay.Format,
													 d3ddmDisplay.Format, D3DFMT_D32)))
		{
			d3dpp.AutoDepthStencilFormat = D3DFMT_D32;
			g_pUtils->Log("Depth buffer depth: 32 bit; no stencil buffer support");
		} else

		if (SUCCEEDED(m_pD3D->CheckDepthStencilMatch(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, d3ddmDisplay.Format,
													 d3ddmDisplay.Format, D3DFMT_D16)))
		{
			d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
			g_pUtils->Log("Depth buffer depth: 16 bit; no stencil buffer support");
		} 
		else throw InitException("Backbuffer format not supported!");
	} 


    if (FULLSCREEN)
    {
	    d3dpp.BackBufferWidth = m_iResX;
	    d3dpp.BackBufferHeight = m_iResY;

	    d3dpp.BackBufferCount = 1;
	    d3dpp.MultiSampleType = D3DMULTISAMPLE_NONE;
		d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	    d3dpp.hDeviceWindow = hWnd;
	    d3dpp.Windowed = false;
	    d3dpp.EnableAutoDepthStencil = true;
	    d3dpp.FullScreen_RefreshRateInHz = /*d3ddmDisplay.RefreshRate;//*/D3DPRESENT_RATE_DEFAULT;
	    d3dpp.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_ONE;
        d3dpp.hDeviceWindow = hWnd;

    } else
    {
        d3dpp.Windowed = TRUE;
        d3dpp.BackBufferCount = 1;
        d3dpp.MultiSampleType = D3DMULTISAMPLE_NONE;
        d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
        d3dpp.EnableAutoDepthStencil = true;   
        d3dpp.hDeviceWindow = hWnd;
    }


    
    hres = m_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
                                 D3DCREATE_HARDWARE_VERTEXPROCESSING,
                                 &d3dpp, &m_pd3dDevice );
    if(SUCCEEDED(hres))
    {
        g_pUtils->Log("Hardware vertexprocessing supported.");
    }
    else
    {
        g_pUtils->Log("Hardware vertexprocessing not supported, switching to software vertexprocessing.");
        hres = m_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
                                     D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                     &d3dpp, &m_pd3dDevice );
        if (FAILED(hres)) 
        {
            SAFE_RELEASE(m_pD3D);
            throw InitException(hres, "CreateDevice FAILED!");
        }
    }

    m_pd3dDevice->GetDeviceCaps(&m_d3dCaps);

	sprintf(msg, "Max. texture blend stages: %i", m_d3dCaps.MaxTextureBlendStages);
	g_pUtils->Log(msg);
	sprintf(msg, "Max. simultaneous textures: %i", m_d3dCaps.MaxSimultaneousTextures);
	g_pUtils->Log(msg);
    g_pUtils->Log("D3D Initialization finished\n");
}

D3DLayer::~D3DLayer()
{
    assert(m_pd3dDevice);
    SAFE_RELEASE(m_pd3dDevice);
    assert(m_pD3D);
    SAFE_RELEASE(m_pD3D);
}

LPDIRECT3DDEVICE8 D3DLayer::Getd3dDevice()
{
    return m_pd3dDevice;
}

D3DCAPS8 D3DLayer::GetD3DCaps()
{
    return m_d3dCaps;
}

void D3DLayer::SaveScreenshot(std::string _fileName)
{
	LPDIRECT3DSURFACE8 pSurfScreen;
	D3DLOCKED_RECT lockedRect;
	D3DDISPLAYMODE dm;
	HRESULT hres;

	// get current resolution
	hres = m_pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &dm);
	assert(SUCCEEDED(hres));

	// create surface to hold image data
	hres = m_pd3dDevice->CreateImageSurface(dm.Width, dm.Height, D3DFMT_A8R8G8B8, &pSurfScreen);
	assert(SUCCEEDED(hres));

	// copy front surface to pSurfScreen
	hres = m_pd3dDevice->GetFrontBuffer(pSurfScreen);
	assert(SUCCEEDED(hres));

	// get pointer to image data
	hres = pSurfScreen->LockRect(&lockedRect, NULL, D3DLOCK_READONLY);
	assert(SUCCEEDED(hres));


// save bitmap to disk
	BITMAPFILEHEADER bfh;
	BITMAPINFOHEADER bih;
	ofstream out;

	// prepare BitmapFileHeader
	bfh.bfType = 'MB';
	bfh.bfSize = sizeof(bfh) + sizeof(bih) + dm.Height*dm.Width*3;
	bfh.bfReserved1 = bfh.bfReserved2 = 0;
	bfh.bfOffBits = sizeof(bfh) + sizeof(bih);

	// prepare BitmapInfoHeader
	bih.biSize = sizeof(bih);
	bih.biWidth = dm.Width;
	bih.biHeight = dm.Height;
	bih.biPlanes = 1;
	bih.biBitCount = 24;
	bih.biCompression = BI_RGB;
	bih.biSizeImage = 0;
	bih.biXPelsPerMeter = 0;
	bih.biYPelsPerMeter = 0;
	bih.biClrUsed = 0;
	bih.biClrImportant = 0;

	// copy bits into buffer to convert into 24 bit bottom-up DIB
	BYTE * pBitsOut = new BYTE[dm.Width*dm.Height*3 + 1];

	DWORD * pSrc = (DWORD*)lockedRect.pBits;
	BYTE * pDest = pBitsOut;

	for (int y = (int)dm.Height-1; y>=0; y--)
	{
		pSrc = (DWORD*)lockedRect.pBits + y * (lockedRect.Pitch>>2);
		for (int x=0; x<(int)dm.Width; x++)
		{
			*(DWORD*)pDest = *pSrc;

			pDest += 3;
			pSrc++;
		}
	}

	// construct filename
	if (_fileName == "")
	{
		_fileName = "screenshot";
	}

	int ind = 0;
	char szComplFileName[128];
	WIN32_FIND_DATA findData;
	do
	{
		sprintf(szComplFileName, "%s%d.bmp", _fileName.c_str(), ind);
		ind++;	
	} while (FindFirstFile(szComplFileName, &findData) != INVALID_HANDLE_VALUE);

	// write file
	out.open(szComplFileName, ios_base::binary);

	out.write((char*)&bfh, sizeof(bfh));
	out.write((char*)&bih, sizeof(bih));
	out.write((char*)pBitsOut, dm.Width*dm.Height*3);

	out.flush();
	out.close();

	delete [] pBitsOut;

	pSurfScreen->UnlockRect();
	pSurfScreen->Release();
}

POINT D3DLayer::GetRes()
{
	POINT pt = { m_iResX, m_iResY };
	return pt;
}

bool D3DLayer::StencilSupported()
{
	return m_bStencil;
}
