// D3DRenderLayer.cpp: Implementierung der Klasse D3DRenderLayer.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "D3DRenderLayer.h"
#include "D3DLayer.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

D3DRenderLayer::D3DRenderLayer(D3DLayer * _d3dLayer)
{
    m_pd3dDevice = _d3dLayer->Getd3dDevice();
    assert(m_pd3dDevice);

    SetInitialRenderStates();
}

D3DRenderLayer::~D3DRenderLayer()
{
//	m_pd3dDevice->SetStreamSource(0,NULL,0);

    unsigned i;
    for (i=0; i<m_vpTexture.size(); i++)
    {
        m_vpTexture[i]->Release();
    }

    for (i=0; i<m_vpVB.size(); i++)
    {
        m_vpVB[i]->Release();
    }

    for (i=0; i<m_vpIB.size(); i++)
    {
        m_vpIB[i]->Release();
    }
}

void D3DRenderLayer::SetInitialRenderStates()
{

    m_pd3dDevice->SetRenderState( D3DRS_SHADEMODE, D3DSHADE_GOURAUD );    
    m_pd3dDevice->SetRenderState( D3DRS_CULLMODE,  D3DCULL_CCW );
    m_pd3dDevice->SetRenderState( D3DRS_FILLMODE,  D3DFILL_SOLID );

    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_MODULATE );
//    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_SELECTARG2 );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER,   D3DTEXF_LINEAR );


	m_pd3dDevice->SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	m_pd3dDevice->SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT);
	m_pd3dDevice->SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_MODULATE2X);
//	m_pd3dDevice->SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_ADDSIGNED);
    m_pd3dDevice->SetTextureStageState( 1, D3DTSS_MAGFILTER,   D3DTEXF_ANISOTROPIC );
    m_pd3dDevice->SetTextureStageState( 1, D3DTSS_MIPFILTER,   D3DTEXF_LINEAR );
    m_pd3dDevice->SetTextureStageState( 1, D3DTSS_MINFILTER,   D3DTEXF_ANISOTROPIC );
/*
	// enable fog
    m_pd3dDevice->SetRenderState( D3DRS_FOGENABLE,  true );
    m_pd3dDevice->SetRenderState( D3DRS_FOGCOLOR,  0xff00ff );
    m_pd3dDevice->SetRenderState( D3DRS_FOGTABLEMODE,     D3DFOG_LINEAR );	
//	m_pd3dDevice->SetRenderState( D3DRS_RANGEFOGENABLE, true );


	float Start = 400.0f;
	float End = 50000.0f;
	float Density = 0.001f;
	
	m_pd3dDevice->SetRenderState(D3DRS_FOGSTART, *(DWORD *)(&Start));
	m_pd3dDevice->SetRenderState(D3DRS_FOGEND,   *(DWORD *)(&End));
    	  
    m_pd3dDevice->SetRenderState(D3DRS_FOGDENSITY, *(DWORD *)(&Density));	 
*/



	// set a default material
    D3DMATERIAL8 mtrl;
    ZeroMemory( &mtrl, sizeof(D3DMATERIAL8) );
    
    mtrl.Diffuse.r = mtrl.Ambient.r = 1.0f;
    mtrl.Diffuse.g = mtrl.Ambient.g = 1.0f;
    mtrl.Diffuse.b = mtrl.Ambient.b = 1.0f;
    mtrl.Diffuse.a = mtrl.Ambient.a = 1.0f;
        
    mtrl.Specular.r = 1.0f;
    mtrl.Specular.g = 1.0f;
    mtrl.Specular.b = 1.0f;
    mtrl.Specular.a = 1.0f;
    mtrl.Power = 30.0f;

	m_pd3dDevice->SetMaterial(&mtrl);
}


HRESULT D3DRenderLayer::Present()
{
    return m_pd3dDevice->Present(NULL, NULL, NULL, NULL);
}

HRESULT D3DRenderLayer::EndScene()
{
    return m_pd3dDevice->EndScene();
}

HRESULT D3DRenderLayer::BeginScene()
{
    return m_pd3dDevice->BeginScene();
}

HRESULT D3DRenderLayer::Clear(D3DCOLOR _color, float _z)
{
    return m_pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER/*|D3DCLEAR_STENCIL*/, _color, _z, 0);
}

handleTex D3DRenderLayer::CreateTexture(const char *_filename)
{
    LPDIRECT3DTEXTURE8 pTex;

	std::string fullname = g_pUtils->GetFullPath(_filename);

    HRESULT hres = D3DXCreateTextureFromFile(m_pd3dDevice, fullname.c_str(), &pTex);
    if (FAILED(hres))
    {
        std::string msg = "Texture Initialization failed. File: ";
        throw InitException(hres, msg + fullname.c_str());
    }

	g_pUtils->Log(std::string("Created texture ") + _filename);

    m_vpTexture.push_back(pTex);

    return m_vpTexture.size()-1;
}

HRESULT D3DRenderLayer::SetTexture(DWORD _stage, handleTex _tex)
{
    return m_pd3dDevice->SetTexture(_stage, m_vpTexture[_tex]);
}

HRESULT D3DRenderLayer::SetTransform(D3DTRANSFORMSTATETYPE _state, CONST D3DMATRIX* _matrix)
{
    return m_pd3dDevice->SetTransform(_state, _matrix);
}


HRESULT D3DRenderLayer::SetMaterial(CONST D3DMATERIAL8* _material)
{
    return m_pd3dDevice->SetMaterial(_material);
}

handleVB D3DRenderLayer::CreateVertexBuffer(UINT _buffersize, size_t _vertexsize, DWORD _usage, DWORD _fvf)
{
	assert (_buffersize <= g_pD3DLayer->GetD3DCaps().MaxVertexIndex);

	LPDIRECT3DVERTEXBUFFER8 pVB;

	// select managed pool if static VB
	_D3DPOOL pool = _usage & D3DUSAGE_DYNAMIC ? D3DPOOL_DEFAULT : D3DPOOL_MANAGED;

	HRESULT hres = m_pd3dDevice->CreateVertexBuffer(_buffersize*_vertexsize, _usage, _fvf, pool, &pVB);
    if (FAILED(hres))
    {
        throw InitException(hres, "VB creation failed.");
    }

	m_vpVB.push_back(pVB);

	return m_vpVB.size() -1;
}

HRESULT D3DRenderLayer::LockVB(handleVB _buffer, UINT _offset, UINT _size, BYTE** _data, DWORD _flags)
{
	return m_vpVB[_buffer]->Lock(_offset, _size, _data, _flags);
}

/**
 *  _data		: pointer to source vertex data
 *	_numVertices: size of source data in vertices
 *  _vertexsize : size of a single vertex structure
 *  _offset		: offset into VB in vertices
 */
void D3DRenderLayer::FillVB(handleVB _buffer, BYTE *_data, int _numVertices, size_t _vertexsize, int _offset)
{
	BYTE * pLocked;
	HRESULT hres = m_vpVB[_buffer]->Lock(_offset*_vertexsize, _numVertices*_vertexsize,
										 &pLocked, 0 );
	assert(SUCCEEDED(hres));

	memcpy(pLocked, _data, _vertexsize * _numVertices);

	m_vpVB[_buffer]->Unlock();
	assert(SUCCEEDED(hres));
}

HRESULT D3DRenderLayer::UnlockVB(handleVB _buffer)
{
	return m_vpVB[_buffer]->Unlock();
}

HRESULT D3DRenderLayer::SetStreamSource(UINT _number,handleVB _data, UINT _stride)
{
	HRESULT ret = m_pd3dDevice->SetStreamSource(_number, m_vpVB[_data], _stride);

	assert(SUCCEEDED(ret));

	return ret;
}	

HRESULT D3DRenderLayer::SetVertexShader(DWORD _handle)
{
	HRESULT ret = m_pd3dDevice->SetVertexShader(_handle);

	assert(SUCCEEDED(ret));

	return ret;
}


HRESULT D3DRenderLayer::DrawPrimitive(D3DPRIMITIVETYPE _type, UINT _start, UINT _count)
{
	HRESULT ret;
	ret = m_pd3dDevice->DrawPrimitive(_type, _start, _count);

	assert(SUCCEEDED(ret));
	return ret;
}


/**
 *  _length : length of IB in indices, WORD-indices assumed
 *  
 */
handleIB D3DRenderLayer::CreateIndexBuffer(UINT _length, DWORD _usage)
{
	LPDIRECT3DINDEXBUFFER8 pIB;

	// select managed pool if static IB
	_D3DPOOL pool = _usage & D3DUSAGE_DYNAMIC ? D3DPOOL_DEFAULT : D3DPOOL_MANAGED;

	HRESULT hres = m_pd3dDevice->CreateIndexBuffer(_length*sizeof(WORD), _usage, D3DFMT_INDEX16, pool, &pIB);
    if (FAILED(hres))
    {
        throw InitException(hres, "IB creation failed.");
    }

	m_vpIB.push_back(pIB);

	return m_vpIB.size()-1;
}

/**
 *	_data       : pointer to index data
 *  _numIndices : number of indices
 */
void D3DRenderLayer::FillIB(handleIB _buffer, const WORD *_data, int _numIndices)
{
    HRESULT hres;
	WORD * pLocked;
	hres = m_vpIB[_buffer]->Lock(0, sizeof(WORD)*_numIndices, (BYTE**)&pLocked, 0);
    assert(SUCCEEDED(hres));

	memcpy(pLocked, _data, sizeof(WORD)*_numIndices);

	hres = m_vpIB[_buffer]->Unlock();
    assert(SUCCEEDED(hres));
}

HRESULT D3DRenderLayer::DrawIndexedPrimitive(D3DPRIMITIVETYPE _type, UINT _minindex, UINT _numvertices, UINT _startindex, UINT _count)
{
    HRESULT ret;
	ret = m_pd3dDevice->DrawIndexedPrimitive(_type, _minindex, _numvertices, _startindex, _count);

	assert(SUCCEEDED(ret));
	return ret;
}

HRESULT D3DRenderLayer::SetIndices(handleIB _buffer, UINT _baseindex)
{
    HRESULT ret;
    ret = m_pd3dDevice->SetIndices(m_vpIB[_buffer], _baseindex);

	assert(SUCCEEDED(ret));
	return ret;
}

HRESULT D3DRenderLayer::SetRenderState(D3DRENDERSTATETYPE _type, DWORD _state)
{
	return m_pd3dDevice->SetRenderState(_type, _state);
}

HRESULT D3DRenderLayer::SetTextureStageState(DWORD _stage, D3DTEXTURESTAGESTATETYPE _type, DWORD _value)
{
	return m_pd3dDevice->SetTextureStageState(_stage, _type, _value);
}


HRESULT D3DRenderLayer::BeginStateBlock()
{
	return m_pd3dDevice->BeginStateBlock();
}

HRESULT D3DRenderLayer::EndStateBlock(DWORD * _token)
{
	return m_pd3dDevice->EndStateBlock(_token);
}

HRESULT D3DRenderLayer::DeleteStateBlock(DWORD _token)
{
	return m_pd3dDevice->DeleteStateBlock(_token);
}

HRESULT D3DRenderLayer::ApplyStateBlock(DWORD _token)
{
	return m_pd3dDevice->ApplyStateBlock(_token);
}

handleTex D3DRenderLayer::CreateTexture(  const char * _filename,
										  UINT _width,
										  UINT _height,
										  UINT _miplevels,
										  DWORD _usage,
										  D3DFORMAT _format,
										  D3DPOOL _pool,
										  DWORD _filter,
										  DWORD _mipfilter,
										  D3DCOLOR _colorkey,
										  D3DXIMAGE_INFO* _srcInfo,
										  PALETTEENTRY* _palette)
{
    LPDIRECT3DTEXTURE8 pTex;

	std::string fullname = g_pUtils->GetFullPath(_filename);

    HRESULT hres = D3DXCreateTextureFromFileEx(m_pd3dDevice, fullname.c_str(), _width, _height,
											   _miplevels, _usage, _format, _pool, _filter, _mipfilter, 
											   _colorkey, _srcInfo, _palette, &pTex);
    if (FAILED(hres))
    {
        std::string msg = "Texture Initialization failed. File: ";
        throw InitException(hres, msg + fullname.c_str());
    }

	g_pUtils->Log(std::string("Created texture ") + _filename);

    m_vpTexture.push_back(pTex);

    return m_vpTexture.size()-1;
}

HRESULT D3DRenderLayer::GetViewport(D3DVIEWPORT8 *_viewport)
{
	return m_pd3dDevice->GetViewport(_viewport);
}

HRESULT D3DRenderLayer::LockRect(handleTex _tex, UINT _level, D3DLOCKED_RECT *_lockedRect, CONST RECT *_rect, DWORD _flags)
{
	return m_vpTexture[_tex]->LockRect(_level, _lockedRect, _rect, _flags);
}

HRESULT D3DRenderLayer::UnlockRect(handleTex _tex, UINT _level)
{
	return m_vpTexture[_tex]->UnlockRect(_level);
}
