// Energy.cpp: Implementierung der Klasse Energy.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Energy.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Energy::Energy()
{
	SETPARAMETERFILE("EnergyParams");

	m_hVB = g_pD3DRenderLayer->CreateVertexBuffer(4, sizeof(ENERGYVERTEX), D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY, D3DFVF_ENERGYVERTEX );

	m_hTex = g_pD3DRenderLayer->CreateTexture(LOADSTRING("m_szTexture", "media\\energy.bmp").c_str());

	m_hSound = g_pDSoundLayer->CreateSound(LOADSTRING("m_szSound", "media\\energy.wav").c_str());

	m_vertQuad[0].u = 1.0f;
	m_vertQuad[0].v = 1.0f;
	m_vertQuad[1].u = 0.0f;
	m_vertQuad[1].v = 1.0f;
	m_vertQuad[2].u = 1.0f;
	m_vertQuad[2].v = 0.0f;
	m_vertQuad[3].u = 0.0f;
	m_vertQuad[3].v = 0.0f;

	for (int i=0; i<4; i++)
	{
		m_pRipple[i].m_fLife = 0.0f;
	}

	m_iNextRipple = 0;
}

Energy::~Energy()
{

}

void Energy::Create(D3DXVECTOR3 & _pos, D3DXVECTOR3 & _n, D3DXVECTOR3 & _up, float _force)
{
	static const D3DXVECTOR3 pQuadVertex[] = { D3DXVECTOR3(0.0f,  1.0f,  1.0f),
											   D3DXVECTOR3(0.0f,  1.0f, -1.0f),
											   D3DXVECTOR3(0.0f, -1.0f,  1.0f), 
											   D3DXVECTOR3(0.0f, -1.0f, -1.0f) };

	D3DXVECTOR3 left;
	D3DXVec3Cross(&left, &_n, &_up);

	D3DXMATRIX trans;

	// x-axis
	*(D3DXVECTOR3*)&trans._11 = _n;

	// y-axis
	*(D3DXVECTOR3*)&trans._21 = _up;

	// z-axis
	*(D3DXVECTOR3*)&trans._31 = left;

	D3DXVec3TransformNormal(&m_pRipple[m_iNextRipple].m_pTransQuadVertex[0], &pQuadVertex[0], &trans);
	D3DXVec3TransformNormal(&m_pRipple[m_iNextRipple].m_pTransQuadVertex[1], &pQuadVertex[1], &trans);
	D3DXVec3TransformNormal(&m_pRipple[m_iNextRipple].m_pTransQuadVertex[2], &pQuadVertex[2], &trans);
	D3DXVec3TransformNormal(&m_pRipple[m_iNextRipple].m_pTransQuadVertex[3], &pQuadVertex[3], &trans);

	m_pRipple[m_iNextRipple].m_fLife = 1.0f;
	m_pRipple[m_iNextRipple].m_fMaxRadius = _force;
	m_pRipple[m_iNextRipple].m_vPos = _pos;

	if (++m_iNextRipple == 4) m_iNextRipple = 0;

	g_pDSoundLayer->PlaySound(m_hSound);
}

void Energy::FrameMove()
{
	for (int i=0; i<4; i++)
	{
		if (m_pRipple[i].m_fLife > 0.0f)
			m_pRipple[i].m_fLife -= 0.01f;
		else
			if (m_pRipple[i].m_fLife < 0.0f) m_pRipple[i].m_fLife = 0.0f;
	}
}

void Energy::Render()
{
	bool first = true;

	for (int i=0; i<4; i++)
	{
		if (m_pRipple[i].m_fLife <= 0.0f) continue;

		if (first)
		{
			first = false;

			g_pD3DRenderLayer->SetTexture(0, m_hTex);

			g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
			g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
			g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);

			g_pD3DRenderLayer->SetRenderState(D3DRS_ALPHABLENDENABLE, true);
			g_pD3DRenderLayer->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCCOLOR);
			g_pD3DRenderLayer->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCCOLOR);

//		    g_pD3DRenderLayer->SetRenderState( D3DRS_FOGENABLE,  false );
			g_pD3DRenderLayer->SetRenderState( D3DRS_CULLMODE,  D3DCULL_NONE);
			g_pD3DRenderLayer->SetRenderState( D3DRS_ZWRITEENABLE,  false);
			g_pD3DRenderLayer->SetRenderState( D3DRS_LIGHTING,  false);

			D3DXMATRIX mat;
			D3DXMatrixIdentity(&mat);
			g_pD3DRenderLayer->SetTransform(D3DTS_WORLD, &mat);

			
			g_pD3DRenderLayer->SetVertexShader(D3DFVF_ENERGYVERTEX);
			g_pD3DRenderLayer->SetStreamSource(0, m_hVB, sizeof(ENERGYVERTEX));
		}

		m_vertQuad[0].pos = m_pRipple[i].m_pTransQuadVertex[0] * ( - m_pRipple[i].m_fLife+1.0f) *
							m_pRipple[i].m_fMaxRadius + m_pRipple[i].m_vPos;
		m_vertQuad[1].pos = m_pRipple[i].m_pTransQuadVertex[1] * ( - m_pRipple[i].m_fLife+1.0f) * 
							m_pRipple[i].m_fMaxRadius + m_pRipple[i].m_vPos;
		m_vertQuad[2].pos = m_pRipple[i].m_pTransQuadVertex[2] * ( - m_pRipple[i].m_fLife+1.0f) * 
							m_pRipple[i].m_fMaxRadius + m_pRipple[i].m_vPos;
		m_vertQuad[3].pos = m_pRipple[i].m_pTransQuadVertex[3] * ( - m_pRipple[i].m_fLife+1.0f) * 
							m_pRipple[i].m_fMaxRadius + m_pRipple[i].m_vPos;

		m_vertQuad[0].c = m_vertQuad[1].c = m_vertQuad[2].c = m_vertQuad[3].c = 
			Math::ColorScale(0xff0000, m_pRipple[i].m_fLife);

		ENERGYVERTEX * pLocked;
		g_pD3DRenderLayer->LockVB(m_hVB, 0, 4*sizeof(ENERGYVERTEX), (BYTE**)&pLocked, D3DLOCK_DISCARD);
		memcpy(pLocked, m_vertQuad, sizeof(ENERGYVERTEX)*4);
		g_pD3DRenderLayer->UnlockVB(m_hVB);
	
		g_pD3DRenderLayer->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);	
	}

	if (!first)
	{
//	    g_pD3DRenderLayer->SetRenderState( D3DRS_FOGENABLE,  true );
		g_pD3DRenderLayer->SetRenderState(D3DRS_ALPHABLENDENABLE, false);
		g_pD3DRenderLayer->SetRenderState( D3DRS_CULLMODE,  D3DCULL_CCW);
		g_pD3DRenderLayer->SetRenderState( D3DRS_ZWRITEENABLE,  true);

	}
}
