// Game.cpp: Implementierung der Klasse Game.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Game.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Game::Game()
{
	SETPARAMETERFILE("GameParams");

	m_fGravity = LOADFLOAT("m_fGravity", 0.4f);
	m_iTargetFPS = LOADINT("m_iTargetFPS", 40);
	m_iMoveSpeed = LOADINT("m_iMoveSpeed", 10);
	m_iMouseSensitivity = LOADINT("m_iMouseSensitivity", 4);
	m_iInvertMouse = LOADINT("m_bInvertMouse", 1) ? -1:1;
	m_fExplosionStrength = LOADFLOAT("m_fExplosionStrength", 150.0f);

	m_iNumExplosionParticles = LOADINT("m_iNumExplosionParticles", 100);

	m_hSoundExplode = g_pDSoundLayer->CreateSound(LOADSTRING("m_szExplosionSound", "media/EXPLODE.wav").c_str());

	m_iTicksSinceLastExplode = 50000;

    // lights
	m_fAmbient = LOADFLOAT("m_fAmbient", 0.3f);
    g_pD3DRenderLayer->SetRenderState( D3DRS_AMBIENT, (int)(m_fAmbient*0xffffff));
 
	
    D3DLIGHT8 light;
    ZeroMemory( &light, sizeof(D3DLIGHT8) );
    light.Type       = D3DLIGHT_DIRECTIONAL;
   
	light.Diffuse.r  = 1.0f;
    light.Diffuse.g  = 1.0f;
    light.Diffuse.b  = 1.0f;
    light.Specular.r = 1.0f;
    light.Specular.g = 1.0f;
    light.Specular.b = 1.0f;
    
    m_vLightDir.x = LOADFLOAT("m_fLightDirX", 0.0f);
    m_vLightDir.y = LOADFLOAT("m_fLightDirY", -1.0f);
    m_vLightDir.z = LOADFLOAT("m_fLightDirZ", 0.0f);

    D3DXVec3Normalize( (D3DXVECTOR3*)&m_vLightDir, &m_vLightDir );
	light.Direction = m_vLightDir;
    g_pD3DLayer->Getd3dDevice()->SetLight( 0, &light );
    g_pD3DLayer->Getd3dDevice()->LightEnable( 0, TRUE );


	m_hVBShadowQuad = g_pD3DRenderLayer->CreateVertexBuffer(4, sizeof(SHADOWQUADVERTEX), 
														    D3DUSAGE_WRITEONLY, D3DFVF_SHADOWQUADVERTEX);
	SHADOWQUADVERTEX quad[4];
	POINT res = g_pD3DLayer->GetRes();
	quad[0].pos = D3DXVECTOR4(0.0f,			 0.0f,		  0.0f, 1.0f);
	quad[1].pos = D3DXVECTOR4((float)res.x,  0.0f,		  0.0f, 1.0f);
	quad[2].pos = D3DXVECTOR4(0.0f,			(float)res.y, 0.0f, 1.0f);
	quad[3].pos = D3DXVECTOR4((float)res.x, (float)res.y, 0.0f, 1.0f);
	
	quad[0].color = 0x7f000000;
	quad[1].color = 0x7f000000;
	quad[2].color = 0x7f000000;
	quad[3].color = 0x7f000000;
	g_pD3DRenderLayer->FillVB(m_hVBShadowQuad, (BYTE*)quad, 4, sizeof(SHADOWQUADVERTEX));
    
	m_pTerrain = NULL;
	m_pSphere = NULL;


	m_bFollow = 1;
	m_vFreeCameraPos = m_Camera.GetPos();

	m_bPaused = false;

	std::string str = LOADSTRING("m_szExplosionTexture", "media/explode.dds");
	m_hEffSmoke = g_pParticleSystem->CreateEffect(str.c_str(), 0.002f, 
												      0x40cccccc, 0x00ffffff, 0.95f,
												      D3DXVECTOR3(-m_fGravity*0.3f, -m_fGravity*0.07f, -m_fGravity*0.2f));
	m_hEffExplosion2 = g_pParticleSystem->CreateEffect(str.c_str(), 0.01f, 
												      0xFFFFFF00, 0x00FFFFFF, 0.95f,
												      D3DXVECTOR3(0.0f, -m_fGravity*0.3f, 0.0f));
	m_hEffExplosion1 = g_pParticleSystem->CreateEffect(str.c_str(), 0.03f, 
												      0xFFFF3333, 0x00FF0022, 1.0f,
												      D3DXVECTOR3(0.0f, -m_fGravity*0.3f, 0.0f));

	memset(m_iMoving, 0, 3*sizeof(int));

    InitGeometry();

	QueryPerformanceFrequency(&m_liTicksPerSecond);
}

Game::~Game()
{
	SAFE_DELETE(m_pSphere);
	SAFE_DELETE(m_pTerrain);
}

void Game::Render()
{
    // Clear the backbuffer to a blue color
    g_pD3DRenderLayer->Clear();

    // Begin the scene
    g_pD3DRenderLayer->BeginScene();
  
	m_pTerrain->Render(m_Camera);

	if (g_pD3DLayer->StencilSupported())
	{
		m_pSphere->RenderShadow();

		RenderShadow();
	}

	m_pSphere->Render();
	
	g_pParticleSystem->Render(m_Camera);
	
	g_pEnergy->Render();

	g_pD3DFont->DrawText(20, 20, 0xffffffff, m_szFPS);

    // End the scene
    g_pD3DRenderLayer->EndScene();

    g_pD3DRenderLayer->Present();

}

void Game::FrameMove(int _ticks)
{
	if (!m_bPaused)
	{
		m_pSphere->FrameMove();
		g_pParticleSystem->FrameMove();
		g_pEnergy->FrameMove();

		m_iTicksSinceLastExplode ++;
	}

	MoveCamera();

	if (m_pSphere->HasStopped()) Explode();

	float fps = (float)m_liTicksPerSecond.QuadPart / _ticks;
	sprintf(m_szFPS, "FPS : %3.0f", fps);
}


/**
 *	0 : x-axis, 1 : y-axis, 2 : z-axis
 */
void Game::Move(int _axis, int _dir)
{
	m_iMoving[_axis] = _dir;
}


void Game::Yaw(float _alpha)
{
	_alpha *= m_iMouseSensitivity;

	m_Camera.Yaw(_alpha);
}

void Game::InitGeometry()
{    
    
	m_pTerrain = new Terrain();
	m_pSphere = new Sphere();

}

void Game::Pitch(float _alpha)
{
	_alpha *= m_iInvertMouse * m_iMouseSensitivity;

	m_Camera.Pitch(_alpha);
}

Terrain* Game::GetTerrain()
{
	return m_pTerrain;
}

void Game::RenderShadow()
{

    // Set renderstates (disable z-buffering, enable stencil, disable fog, and
    // turn on alphablending)
    g_pD3DRenderLayer->SetRenderState( D3DRS_ZENABLE,          FALSE );
    g_pD3DRenderLayer->SetRenderState( D3DRS_STENCILENABLE,    TRUE );
//    g_pD3DRenderLayer->SetRenderState( D3DRS_FOGENABLE,  false );
    g_pD3DRenderLayer->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
    g_pD3DRenderLayer->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_SRCALPHA );
    g_pD3DRenderLayer->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );

    g_pD3DRenderLayer->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_DIFFUSE );
    g_pD3DRenderLayer->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
    g_pD3DRenderLayer->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_DIFFUSE );
    g_pD3DRenderLayer->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);

    // Only write where stencil val >= 1 (count indicates # of shadows that
    // overlap that pixel)
    g_pD3DRenderLayer->SetRenderState( D3DRS_STENCILREF,  0x1 );
    g_pD3DRenderLayer->SetRenderState( D3DRS_STENCILFUNC, D3DCMP_LESSEQUAL );
    g_pD3DRenderLayer->SetRenderState( D3DRS_STENCILPASS, D3DSTENCILOP_ZERO );

	
    // Draw a big, gray square
    g_pD3DRenderLayer->SetVertexShader( D3DFVF_SHADOWQUADVERTEX );
    g_pD3DRenderLayer->SetStreamSource( 0, m_hVBShadowQuad, sizeof(SHADOWQUADVERTEX) );
    g_pD3DRenderLayer->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 );

    // Restore render states
    g_pD3DRenderLayer->SetRenderState( D3DRS_ZENABLE,          TRUE );
    g_pD3DRenderLayer->SetRenderState( D3DRS_STENCILENABLE,    FALSE );
    g_pD3DRenderLayer->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
//    g_pD3DRenderLayer->SetRenderState( D3DRS_FOGENABLE,  true );

}

void Game::ToggleFollow()
{
	m_bFollow ^= 1;

	if (m_bFollow) m_vFreeCameraPos = m_Camera.GetPos();
	else m_Camera.SetPos(m_vFreeCameraPos);
}

Camera Game::GetCamera()
{
	return m_Camera;
}

void Game::Explode()
{
	if (m_iTicksSinceLastExplode < 300) return;
	m_iTicksSinceLastExplode = 0;

	D3DXVECTOR3 vel = m_pSphere->GetVelocity();
	vel.x += m_fExplosionStrength*0.1f*(float)(rand() - (RAND_MAX>>1))/RAND_MAX;
	vel.y += m_fExplosionStrength;
	vel.z += m_fExplosionStrength*0.1f*(float)(rand() - (RAND_MAX>>1))/RAND_MAX;
	m_pSphere->SetVelocity(vel);

	m_pSphere->SetSoundPosition();
	g_pDSoundLayer->PlaySound(m_hSoundExplode);

	float r = m_pSphere->GetRadius();
	D3DXVECTOR3 pos = m_pSphere->GetPos();
	pos.y -= r;
	g_pParticleSystem->CreateParticles(m_hEffSmoke, pos, D3DXVECTOR3(0.0f, 0.0f, 0.0f), 
										40.0f, m_iNumExplosionParticles);
	g_pParticleSystem->CreateParticles(m_hEffExplosion1, pos, D3DXVECTOR3(0.0f, 0.0f, 0.0f), 
										40.0f, m_iNumExplosionParticles);
	g_pParticleSystem->CreateParticles(m_hEffExplosion2, pos, D3DXVECTOR3(0.0f, 0.0f, 0.0f), 
										40.0f, m_iNumExplosionParticles);
}

void Game::TogglePause()
{
	m_bPaused ^= 1;
}

int Game::GetTargetFPS()
{
	return m_iTargetFPS;
}

float Game::GetGravity()
{
	return m_fGravity;
}

D3DXVECTOR3 Game::GetLightDir()
{
	return m_vLightDir;
}

float Game::GetAmbient()
{
	return m_fAmbient;
}

void Game::MoveCamera()
{
	// Move the camera according to user controls
	for (int i=0; i<3; i++)
	{
		if (m_iMoving[i])
		{
			m_Camera.Move(i, (float)m_iMoving[i]*m_iMoveSpeed);
		}
	}

	// set camera position if in following mode
	static D3DXVECTOR3 targetLookAt = m_pSphere->GetPos();
	D3DXVECTOR3 pos, targetPos, d;
	if (m_bFollow)
	{
		
		pos = m_Camera.GetPos();

		d = pos - m_pSphere->GetPos();
		D3DXVec3Normalize(&d, &d);
		targetPos = m_pSphere->GetPos() + 
			d * (D3DXVec3Length(&m_pSphere->GetVelocity())*m_pSphere->GetRadius()+5.0f*m_pSphere->GetRadius());

		targetPos.y *= 0.85f;

		if (!m_pSphere->HasStopped()) 
		{
			pos += 0.005f*(targetPos-pos);
			targetLookAt += 0.05f*(m_pSphere->GetPos() - targetLookAt);
		}


		m_Camera.SetPos(pos);
	}


	// Collide camera-terrain
	float h = m_pTerrain->GetHeightAt(m_Camera.m_vPos.x, m_Camera.m_vPos.z);
	if ( m_Camera.m_vPos.y - m_Camera.m_fCameraRadius < h)
		m_Camera.m_vPos.y = h + m_Camera.m_fCameraRadius;



	if ( m_bFollow)
		m_Camera.SetTransformLookAt(targetLookAt);
	else
		m_Camera.SetTransform();
}
