// MeshFactory.cpp: Implementierung der Klasse MeshFactory.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MeshFactory.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////


float MeshFactory::ms_fArg1;
float MeshFactory::ms_fArg2;

// used in (x,y)-plot
plotfun MeshFactory::ms_pfSimplePlot = NULL;


MeshFactory::MeshFactory()
{
    SetRes(20,20);
}

void MeshFactory::Plot(PLOTPARAMS &pParams, float (__cdecl *f)(float,float),std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex)
{
    ms_pfSimplePlot = f;
    Plot(pParams, PlotSimple, pVertex, pIndex);
}

void MeshFactory::Plot(PLOTPARAMS &pParams, D3DXVECTOR3 (__cdecl *f)(float,float),
                       std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex)
{
    int u,v;

    if (pParams.texu == 0)
        pParams.texu = (float)(pParams.resu - 1);
    if (pParams.texv == 0)
        pParams.texv = float(pParams.resv - 1);

    // fill vertex array with function plot
    int iNumVertices = pParams.resu*pParams.resv;
    *pVertex = std::vector<MESHFACTORYVERTEX>(iNumVertices);
    std::vector<MESHFACTORYVERTEX>::iterator itCurVertex = pVertex->begin();

    float dx = (pParams.umax - pParams.umin) / (pParams.resu-1);
    float dy = (pParams.vmax - pParams.vmin) / (pParams.resv-1);
    float x = pParams.umin;
    float y = pParams.vmin;
    float tu=0;
    float tv=0;
    for (u = 0; u<pParams.resu; u++)
    {
        for (v = 0; v<pParams.resv; v++)
        {
            D3DXVECTOR3 vRes = f(x,y);
            
            itCurVertex->pos.x = pParams.scalex * vRes.x;
            itCurVertex->pos.y = pParams.scaley * vRes.y;
            itCurVertex->pos.z = pParams.scalez * vRes.z;

            itCurVertex->tu = u / (float)pParams.resu * pParams.texu;
            itCurVertex->tv = v / (float)pParams.resv * pParams.texv;
            /*
            itCurVertex->tu = tu;
            itCurVertex->tv = tv;
            */
            itCurVertex++;

            y+=dy;
            tv += dy/(pParams.vmax-pParams.vmin);
        }
        x+=dx;
        tu += dx/(pParams.umax-pParams.umin);
        y=pParams.vmin;
        tv=0;
    }

    CreateMeshFromVertexgrid(pParams, pVertex, pIndex);
}


void MeshFactory::CreateTorus(float fOuterRadius, float fInnerRadius,std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex)
{
    ms_fArg1 = fOuterRadius;
    ms_fArg2 = fInnerRadius;

    PLOTPARAMS pparams;
    pparams.resu = m_iResU;
    pparams.resv = m_iResV;
    pparams.texu = 1.0f;
    pparams.texv = 1.0f;
    pparams.scalex = 1.0f;
    pparams.scaley = 1.0f;
    pparams.scalez = 1.0f;
    pparams.umax = D3DX_PI;
    pparams.umin = -D3DX_PI;
    pparams.vmax = D3DX_PI;
    pparams.vmin = -D3DX_PI;
    pparams.uclosed = true;
    pparams.vclosed = true;

    Plot(pparams, &MeshFactory::PlotTorus, pVertex, pIndex);
}


D3DXVECTOR3 MeshFactory::PlotTorus(float u, float v)
{
    D3DXVECTOR3 res;
    res.x = (ms_fArg1-cosf(v)*ms_fArg2)*cosf(u);
    res.y = (ms_fArg1-cosf(v)*ms_fArg2)*sinf(u);
    res.z = sinf(v)*ms_fArg2;
    return res;
}

void MeshFactory::CreateCylinder(float fHeight, float fRadius, std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex)
{
    ms_fArg1 = fRadius;

    PLOTPARAMS pparams;
    pparams.resu = m_iResU;
    pparams.resv = m_iResV;
    pparams.texu = 1.0f;
    pparams.texv = 1.0f;
    pparams.scalex = 1.0f;
    pparams.scaley = 1.0f;
    pparams.scalez = 1.0f;
    pparams.umax = D3DX_PI;
    pparams.umin = -D3DX_PI;
    pparams.vmax = fHeight;
    pparams.vmin = 0;
    pparams.uclosed = true;
    pparams.vclosed = true;

    Plot(pparams, &MeshFactory::PlotCylinder, pVertex, pIndex);
}

D3DXVECTOR3 MeshFactory::PlotCylinder(float u, float v)
{
    D3DXVECTOR3 res;

    res.x = ms_fArg1*cosf(u);
    res.y = v;
    res.z = ms_fArg1*sinf(u);

    return res;
}

/*
 * Sets the resolution for non-user specified plots, e.g. torus, cylinder...
 */
void MeshFactory::SetRes(int iResU, int iResV)
{
    m_iResU = iResU;
    m_iResV = iResV;
}

D3DXVECTOR3 MeshFactory::PlotSimple(float u, float v)
{
    D3DXVECTOR3 res;

    res.x = u;
    res.y = v;
    res.z = ms_pfSimplePlot(u,v);

    return res;
}

void MeshFactory::CreateIcosahedron(float fRadius, std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex)
{
    
    // build icosahedron

    // height of triangles
    float fHeight = sqrtf(3.0f/4.0f)*fRadius;

    // create vertices
    *pVertex = std::vector<MESHFACTORYVERTEX>(12);
    std::vector<MESHFACTORYVERTEX>::iterator itCurVertex = pVertex->begin();

    itCurVertex->pos = D3DXVECTOR3(0.0f, fRadius, 0.0f); // north pole
    D3DXVec3Normalize(&itCurVertex->n, &itCurVertex->pos);
    itCurVertex->tu  = 0;
    itCurVertex->tv  = 0;
    itCurVertex++;
    for (int i=0; i<5;i++) 
    {
        float alpha = i * 2.0f* D3DX_PI / 5.0f;
        itCurVertex->pos = D3DXVECTOR3(cosf(alpha)*fHeight, fRadius/2.0f, sinf(alpha)*fHeight);
        D3DXVec3Normalize(&itCurVertex->n, &itCurVertex->pos);
        itCurVertex->tu  = 1.0f / 3.0f;
        itCurVertex->tv  = i / 5.0f;
        itCurVertex++;
    }

    for (i=0; i<5;i++) 
    {
        float alpha = D3DX_PI/5.0f + i * 2.0f * D3DX_PI / 5.0f;
        itCurVertex->pos = D3DXVECTOR3(cosf(alpha)*fHeight, -fRadius/2, sinf(alpha)*fHeight); 
        D3DXVec3Normalize(&itCurVertex->n, &itCurVertex->pos);
        itCurVertex->tu  = 2.0f / 3.0f;
        itCurVertex->tv  = i / 5.0f;
        itCurVertex++;
    }
    itCurVertex->pos = D3DXVECTOR3(0.0f, -fRadius, 0.0f); // south pole 
    D3DXVec3Normalize(&itCurVertex->n, &itCurVertex->pos);
    itCurVertex->tu  = 1.0;
    itCurVertex->tv  = 0;

    
    // create indices  ( clockwise )

    *pIndex = std::vector<WORD>(20*3);
    std::vector<WORD>::iterator itCurIndex = pIndex->begin();

    for (i=1; i<6; i++)
    {
        *itCurIndex++ = 0; 
        *itCurIndex++ = i==5 ? 1 : i+1; 
        *itCurIndex++ = i;
    }
    
    for (i=1; i<6; i++)
    {
        *itCurIndex++ = i; 
        *itCurIndex++ = i==5 ? 1 : i+1;
        *itCurIndex++ = i+5;

        *itCurIndex++ = i; 
        *itCurIndex++ = i+5;
        *itCurIndex++ = i==1 ? 10 : i+4;
    }
    
    for (i=6; i<11; i++)
    {
        *itCurIndex++ = i; 
        *itCurIndex++ = i==10 ? 6 : i+1;
        *itCurIndex++ = 11;
    }

}

void MeshFactory::CreateGeosphere(int iDepth, float fRadius, std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex)
{
    
    // create Base model
    CreateIcosahedron(fRadius, pVertex, pIndex);

    // get base model dimensions
    int iNumVertices = pVertex->size();
    int iNumTriangles = pIndex->size() /3;


    // calculate dimensions of finished sphere
    int iNewNumVertices  = iNumVertices + iNumTriangles * ((int)pow(4, iDepth)-1) / 2;
    int iNewNumTriangles = (int)pow(4, iDepth) * iNumTriangles;

    // resize vectors to hold finished sphere
    pVertex->resize(iNewNumVertices);
    pIndex->resize(iNewNumTriangles * 3);


    // the indices of inserted vertices are stored in tEdge::vc, which in turn are stored
    // in a list in the vertices with lower index. tEdge::v is the index of the other vertex of that edge
    typedef std::list<Edge> EdgeList;
    std::vector<EdgeList> vEdge = std::vector<EdgeList> (iNewNumVertices);

    for (int depth = 0; depth<iDepth;depth++)
    {

        //traverse edges, init data structures
        int iNumCreatedVertices = 0;
        int iNumCreatedTriangles = 0;
        int pNewVertexIndex[3];
        for (int tri=0; tri<iNumTriangles; tri++)
        {
            for (int v1=0;v1<3;v1++)
            {
                int v2 = (v1 == 2) ? 0 : v1+1;

                // get two indexes for edge, store lower one in index1
                WORD index1, index2;
                index1=(*pIndex)[tri*3+v1];
                index2=(*pIndex)[tri*3+v2];
                if (index1>index2) {int tmp=index2;index2=index1;index1=tmp;}

                // check edgelist of lower vertex for edge
                bool bEdgeExists = false;
                std::list<Edge>::iterator itEdge = vEdge[index1].begin();
                while (itEdge != vEdge[index1].end())
                {
                    // found edge
                    if(itEdge->v == index2)
                    {
                        bEdgeExists = true;
                        pNewVertexIndex[v1] = itEdge->vc;
                        break;
                    }
                    itEdge++;
                }

                // create the edge
                if (!bEdgeExists)
                {
                    Edge e = { index2, iNumVertices + iNumCreatedVertices };
                    vEdge[index1].push_back(e);
                    pNewVertexIndex[v1] = e.vc;
            
                    // store the new vertex
                    MESHFACTORYVERTEX & nv = (*pVertex)[iNumVertices + iNumCreatedVertices];
                    nv.pos = ((*pVertex)[index1].pos + (*pVertex)[index2].pos) / 2.0f;
                    D3DXVec3Normalize(&nv.n, &nv.pos);
                    nv.pos = nv.n * fRadius;

                    iNumCreatedVertices++;

                }
            }

            // new vertices have been found and/or created, now update indices to split tri into four

            (*pIndex)[iNumTriangles*3 + iNumCreatedTriangles*3  ] = (*pIndex)[tri*3  ];
            (*pIndex)[iNumTriangles*3 + iNumCreatedTriangles*3+1] = pNewVertexIndex[0];
            (*pIndex)[iNumTriangles*3 + iNumCreatedTriangles*3+2] = pNewVertexIndex[2];
            iNumCreatedTriangles++;

            (*pIndex)[iNumTriangles*3 + iNumCreatedTriangles*3  ] = (*pIndex)[tri*3+1];
            (*pIndex)[iNumTriangles*3 + iNumCreatedTriangles*3+1] = pNewVertexIndex[1];
            (*pIndex)[iNumTriangles*3 + iNumCreatedTriangles*3+2] = pNewVertexIndex[0];
            iNumCreatedTriangles++;

            (*pIndex)[iNumTriangles*3 + iNumCreatedTriangles*3  ] = (*pIndex)[tri*3+2];
            (*pIndex)[iNumTriangles*3 + iNumCreatedTriangles*3+1] = pNewVertexIndex[2];
            (*pIndex)[iNumTriangles*3 + iNumCreatedTriangles*3+2] = pNewVertexIndex[1];
            iNumCreatedTriangles++;

            // old tri becomes new center tri
            (*pIndex)[tri*3  ] = pNewVertexIndex[0];
            (*pIndex)[tri*3+1] = pNewVertexIndex[1];
            (*pIndex)[tri*3+2] = pNewVertexIndex[2];
        }

        // reset data structures for next pass
        for (int i=0; i<iNumVertices; i++)
        {
            vEdge[i] = std::list<Edge>(0);
        }

        iNumTriangles += iNumCreatedTriangles;
        iNumVertices += iNumCreatedVertices;
    }

    //calc tex coords
    for (int i=0; i<iNumVertices; i++)
    {
        MESHFACTORYVERTEX * v = &(*pVertex)[i];

        float d = sqrtf(v->pos.z * v->pos.z + v->pos.x * v->pos.x);


        float alpha = asinf(v->pos.z / d);

        if (v->pos.x < 0 && v->pos.z > 0) alpha = D3DX_PI - alpha;
        else if (v->pos.x < 0 && v->pos.z < 0) alpha = D3DX_PI - alpha;
        else if (v->pos.x > 0 && v->pos.z < 0) alpha = 2.0f*D3DX_PI + alpha;

        v->tu = d==0?0.5f : alpha / (2.0f*D3DX_PI);

        v->tv = -atanf(v->pos.y / d) / (D3DX_PI) + 0.5f;
    }

}

void MeshFactory::CreateTetrahedron(float fRadius, std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex)
{   

    // create vertices
    *pVertex = std::vector<MESHFACTORYVERTEX>(4);
    std::vector<MESHFACTORYVERTEX>::iterator itCurVertex = pVertex->begin();

    itCurVertex->pos = D3DXVECTOR3(0.0f, fRadius, 0.0f); // north pole
    D3DXVec3Normalize(&itCurVertex->n, &itCurVertex->pos);
    itCurVertex->tu  = 0;
    itCurVertex->tv  = 0;
    itCurVertex++;
    for (int i=0; i<3;i++) 
    {
        float alpha = i * 2.0f* D3DX_PI / 3.0f;
        itCurVertex->pos = D3DXVECTOR3(fRadius*sqrtf(0.75f)*cosf(alpha), -fRadius*0.5f, fRadius*sqrtf(0.75f)*sinf(alpha));
        D3DXVec3Normalize(&itCurVertex->n, &itCurVertex->pos);
        itCurVertex->tu  = 1.0f / 3.0f;
        itCurVertex->tv  = i / 3.0f;
        itCurVertex++;
    }

    // create indices  ( clockwise )

    *pIndex = std::vector<WORD>(4*3);
    std::vector<WORD>::iterator itCurIndex = pIndex->begin();

    *itCurIndex++ = 0; 
    *itCurIndex++ = 2;
    *itCurIndex++ = 1;

    *itCurIndex++ = 0; 
    *itCurIndex++ = 3;
    *itCurIndex++ = 2;

    *itCurIndex++ = 0; 
    *itCurIndex++ = 1;
    *itCurIndex++ = 3;

    *itCurIndex++ = 3; 
    *itCurIndex++ = 1;
    *itCurIndex++ = 2;

}

/*
 * heightfield must have dimensions pParams.resu x pParamas.resv
 */
void MeshFactory::CreateMeshFromHeightfield(const PLOTPARAMS &pParams, BYTE * pHeightfield, std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex)
{
    *pVertex = std::vector<MESHFACTORYVERTEX>(pParams.resu * pParams.resv);

    for(int x=0; x<pParams.resu; x++)
    {
        for (int y=0; y<pParams.resv; y++)
        {
            MESHFACTORYVERTEX * v = &(*pVertex)[y + pParams.resv*x];
            v->pos.x = x*pParams.scalex;
            v->pos.y = y*pParams.scaley;
            v->pos.z = (float)pHeightfield[y + pParams.resv*x] * pParams.scalez;
            v->tu = x / (float)pParams.resu * pParams.texu;
            v->tv = y / (float)pParams.resv * pParams.texv;
        }
    }

    CreateMeshFromVertexgrid(pParams, pVertex, pIndex);
}

/*
 * Vertex grid must have resolution specified in pParams.resu / resv. 
 */
void MeshFactory::CreateMeshFromVertexgrid(const PLOTPARAMS &pParams, std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex)
{
    int u,v;

    int iNumIndices = (pParams.resu-1)*(pParams.resv-1)*2*3;    
    int iNumVertices = pVertex->size();

    // temporary storage for vertex normals used for averaging
    struct VERTEXNORMAL
    {
        float iNumNormals;
        D3DXVECTOR3 vNormal;
    };
    VERTEXNORMAL * pNormals = new VERTEXNORMAL[iNumVertices];
    ZeroMemory(pNormals, sizeof(VERTEXNORMAL)*iNumVertices);

    *pIndex = std::vector<WORD>(iNumIndices);
    std::vector<WORD>::iterator itCurIndex = pIndex->begin();
    D3DXVECTOR3 vNormal;
    int index1, index2, index3;
    for (u=0; u<pParams.resu-1;u++)
    {
        for (v=0; v<pParams.resv-1;v++)
        {
            index1 = u*(pParams.resv) + v;
            index2 = u*(pParams.resv) + v + 1;
            index3 = u*(pParams.resv) + v + pParams.resv;

            // calc normal of triangle
            D3DXVec3Cross(&vNormal, &((*pVertex)[index1].pos - (*pVertex)[index2].pos), &((*pVertex)[index1].pos - (*pVertex)[index3].pos));

            // add normal to 3 vertices of triangle
            pNormals[index1].vNormal = (pNormals[index1].vNormal * pNormals[index1].iNumNormals + vNormal) / 
                                            ++pNormals[index1].iNumNormals;
            pNormals[index2].vNormal = (pNormals[index2].vNormal * pNormals[index2].iNumNormals + vNormal) / 
                                            ++pNormals[index2].iNumNormals;
            pNormals[index3].vNormal = (pNormals[index3].vNormal * pNormals[index3].iNumNormals + vNormal) / 
                                            ++pNormals[index3].iNumNormals;

            // build index buffer
            *itCurIndex++ = index1;
            *itCurIndex++ = index2;
            *itCurIndex++ = index3;


            index1 = u*(pParams.resv) + v + 1;
            index2 = u*(pParams.resv) + v + pParams.resv + 1;
            index3 = u*(pParams.resv) + v + pParams.resv;

            // calc normal of triangle
            D3DXVec3Cross(&vNormal, &((*pVertex)[index1].pos - (*pVertex)[index2].pos), &((*pVertex)[index1].pos - (*pVertex)[index3].pos));

            // add normal to 3 vertices of triangle
            pNormals[index1].vNormal = (pNormals[index1].vNormal * pNormals[index1].iNumNormals + vNormal) / 
                                            ++pNormals[index1].iNumNormals;
            pNormals[index2].vNormal = (pNormals[index2].vNormal * pNormals[index2].iNumNormals + vNormal) / 
                                            ++pNormals[index2].iNumNormals;
            pNormals[index3].vNormal = (pNormals[index3].vNormal * pNormals[index3].iNumNormals + vNormal) / 
                                            ++pNormals[index3].iNumNormals;
            
            // build index buffer
            *itCurIndex++ = index1;
            *itCurIndex++ = index2;
            *itCurIndex++ = index3;
        }
    }


    // smooth normals if structure wraps around
    if (pParams.uclosed)
    {
        for(v=0; v<pParams.resv; v++)
        {
            D3DXVECTOR3 * v1, *v2;
            v1 = &pNormals[v].vNormal;
            v2 = &pNormals[(pParams.resu-1)*pParams.resv +v].vNormal;
            *v1 += *v2;
            *v1 /= 2;
            *v2 = *v1;
        }
    }
    if (pParams.vclosed)
    {
        for(u=0; u<pParams.resu; u++)
        {
            D3DXVECTOR3 * v1, *v2;
            v1 = &pNormals[u*(pParams.resv)].vNormal;
            v2 = &pNormals[u*(pParams.resv) + pParams.resv - 1].vNormal;
            *v1 += *v2;
            *v1 /= 2;
            *v2 = *v1;
        }
    }

    // store normalized normals in vertices
    for (u=0; u<iNumVertices;u++)
    {
        MESHFACTORYVERTEX * v = &(*pVertex)[u];
        D3DXVec3Normalize(&(*pVertex)[u].n, &pNormals[u].vNormal);
    }

    SAFE_DELETE_ARRAY(pNormals);
}

void MeshFactory::CreateMeshFromBitmap(PLOTPARAMS &pParams, char * szBitmapName, std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex)
{
    BITMAPFILEHEADER bfh;
    BITMAPINFOHEADER bih;

    ifstream in(szBitmapName, ios_base::binary);
    if (in.fail())
    {
        std::string msg = "Unable to load heightfield bitmap! (";
        msg += szBitmapName;
        msg += ")";
        throw InitException(msg);
    }

    in.read((char*)&bfh, sizeof(bfh));
    in.read((char*)&bih, sizeof(bih));
    if (bih.biCompression != 0)
    {
        std::string msg = "bitmap compression not supported! (";
        msg += szBitmapName;
        msg += ")";
        throw InitException(msg);
    }
    if (bih.biBitCount != 8)
    {
        std::string msg = "Only 8 bit heightfield supported! (";
        msg += szBitmapName;
        msg += ")";
        throw InitException(msg);
    }

    //store width & height
    pParams.resu = bih.biWidth;
    pParams.resv = bih.biHeight;

    // apply texture once per quad if zero
    if (pParams.texu == 0)
        pParams.texu = (float)(pParams.resu);
    if (pParams.texv == 0)
        pParams.texv = float(pParams.resv);


    // ignore palette & stuff
    in.ignore(bfh.bfOffBits - sizeof(bfh) - sizeof(bih));

    BYTE * pHeightmap = new BYTE[bih.biHeight * bih.biWidth];
	BYTE * pLoc = pHeightmap;
	int border = bih.biWidth % 4;	// bitmap is DWORD-aligned
	if (border) border=4-border;

	for (int h = 0; h<bih.biHeight; h++)
	{
	    in.read((char*)pLoc, bih.biWidth);
		in.ignore(border);
		pLoc += bih.biWidth;
	}



    CreateMeshFromHeightfield(pParams, pHeightmap, pVertex, pIndex);

    delete [] pHeightmap;
}
