// MeshFactory.h: Schnittstelle fr die Klasse MeshFactory.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_MESHFACTORY_H__DB0BAAA6_D99F_11D5_ADB1_0020E061A328__INCLUDED_)
#define AFX_MESHFACTORY_H__DB0BAAA6_D99F_11D5_ADB1_0020E061A328__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#pragma warning (disable : 4786) //"Bezeichner wurde auf '255' Zeichen in den Debug-Informationen verkrzt"

#include "globalsextern.h"
#include <fstream.h>

typedef float (*plotfun)(float,float);

struct PLOTPARAMS
{
    float umin, umax, vmin, vmax;
    float texu, texv;
    int resu, resv;
    float scalex, scaley, scalez;
    bool uclosed, vclosed;
};
struct MESHFACTORYVERTEX
{
    D3DXVECTOR3 pos;
    D3DXVECTOR3 n;
    float tu, tv;
};



#define g_pMeshFactory MeshFactory::GetSingleton()
class MeshFactory:public Singleton<MeshFactory>
{
public:
	MeshFactory();

	void CreateMeshFromBitmap(PLOTPARAMS &pParams, char * szBitmapName, std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex);
	void CreateMeshFromHeightfield(const PLOTPARAMS &pParams, BYTE * pHeightfield, std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex);
	void CreateTetrahedron(float fRadius, std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex);
	void CreateGeosphere(int iDepth, float fRadius, std::vector<MESHFACTORYVERTEX>*pVertexOut, std::vector<WORD>*pIndexOut);
	void CreateIcosahedron(float fRadius, std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex);
    void CreateCylinder(float fHeight, float fRadius, std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex);
	void CreateTorus(float fOuterRadius, float fInnerRadius,std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex);

    void Plot(PLOTPARAMS &pParams, D3DXVECTOR3 (__cdecl *f)(float,float),std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex);
    void Plot(PLOTPARAMS &pParams, float (__cdecl *f)(float,float),std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex);

	void SetRes(int iResU, int iResV);

private:
	int m_iResU;
	int m_iResV;

	void CreateMeshFromVertexgrid(const PLOTPARAMS &pParams, std::vector<MESHFACTORYVERTEX>*pVertex, std::vector<WORD>*pIndex);

	static D3DXVECTOR3 PlotSimple(float u, float v);
	static D3DXVECTOR3 PlotCylinder(float u, float v);
	static D3DXVECTOR3 PlotTorus(float u, float v);

	static plotfun ms_pfSimplePlot;
	static float ms_fArg1;
	static float ms_fArg2;

    struct Edge
    {
        int v;
        int vc;
    };
};

#endif // !defined(AFX_MESHFACTORY_H__DB0BAAA6_D99F_11D5_ADB1_0020E061A328__INCLUDED_)
