// MeshManager.cpp: Implementierung der Klasse MeshManager.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MeshManager.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

MeshManager::MeshManager()
{

    // Create Verterx Buffer with maximum number of vertices
    D3DCAPS8 caps = g_pD3DLayer->GetD3DCaps();

	m_hVB = g_pD3DRenderLayer->CreateVertexBuffer(caps.MaxVertexIndex, sizeof(OBJECTVERTEX), D3DUSAGE_WRITEONLY, D3DFVF_OBJECTVERTEX);

    Mesh::SetVB(m_hVB);
    Mesh::SetFVF(D3DFVF_OBJECTVERTEX, sizeof(OBJECTVERTEX));

    // initialize vector of mesh pointers
    m_pMesh = std::vector<Mesh*>(50);

    m_iNumTotalVertices = 0;
    m_iNumMeshes = 0;
}

MeshManager::~MeshManager()
{
    std::vector<Mesh*>::iterator mi = m_pMesh.begin();
    while (mi != m_pMesh.end())
    {
        delete *mi;
        mi++;
    }
}

Mesh* MeshManager::GetMesh(int m)
{
    assert(m < m_iNumMeshes)

    return m_pMesh[m];
}

Mesh* MeshManager::CreateMesh(const OBJECTVERTEX*pVertices, int iNumVertices, const WORD*pIndices, int iNumIndices)
{
	assert (m_iNumTotalVertices + iNumVertices < g_pD3DLayer->GetD3DCaps().MaxVertexIndex);

    Mesh * pNewMesh = new Mesh(pVertices, iNumVertices, pIndices, iNumIndices, m_iNumTotalVertices);
    m_pMesh[m_iNumMeshes++] = pNewMesh;
    m_iNumTotalVertices += iNumVertices;


    return pNewMesh;
}

Mesh* MeshManager::CreateMesh(const std::vector<MESHFACTORYVERTEX>&vVertex, const std::vector<WORD>&vIndex)
{
    return CreateMesh((OBJECTVERTEX*)&vVertex[0], vVertex.size(), &vIndex[0], vIndex.size());
}
