// Model.cpp: Implementierung der Klasse Model.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Model.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Model::Model()
{
    m_vMaterial = std::vector<D3DMATERIAL8>(0);
    m_vhTexture = std::vector<handleTex>(0);
    m_vpMesh = std::vector<Mesh*>(0);
    m_vmatMesh = std::vector<D3DXMATRIX>(0);

    m_vpChild = std::vector<Model*>(0);
    m_vmatChild = std::vector<D3DXMATRIX>(0);

    m_iNumMeshes = 0;
    m_iNumChildren = 0;
}

Model::~Model()
{
}

const D3DXMATRIX&Model::GetMeshMatrix(int m)
{
    return m_vmatMesh[m];
}

int Model::GetNumMeshes()
{
    return m_iNumMeshes;
}

void Model::AddMesh(Mesh *pMesh, const handleTex _hTex, const D3DMATERIAL8& mat, const D3DXMATRIX& matMesh)
{
    m_vpMesh.resize(m_iNumMeshes+1);
    m_vhTexture.resize(m_iNumMeshes+1);
    m_vMaterial.resize(m_iNumMeshes+1);
    m_vmatMesh.resize(m_iNumMeshes+1);

    m_vpMesh[m_iNumMeshes] = pMesh;
    m_vhTexture[m_iNumMeshes] = _hTex;
    m_vMaterial[m_iNumMeshes] = mat;
    m_vmatMesh[m_iNumMeshes] = matMesh;

    m_iNumMeshes++;
}

Object* Model::GetInstance()
{
    return new Object(this);
}

void Model::Render(const D3DXMATRIX &matWorld)
{
    for(int i=0; i<m_iNumMeshes; i++)
    {
        g_pD3DRenderLayer->SetTransform(D3DTS_WORLD, &(m_vmatMesh[i]*matWorld));
        g_pD3DRenderLayer->SetMaterial(&m_vMaterial[i]);
        g_pD3DRenderLayer->SetTexture(0, m_vhTexture[i]);

        m_vpMesh[i]->Render();
    }
}

void Model::AddChild(Model *pChild, const D3DXMATRIX&matChild)
{
    m_vpChild.resize(m_iNumChildren+1);
    m_vmatChild.resize(m_iNumChildren+1);

    m_vpChild[m_iNumChildren] = pChild;
    m_vmatChild[m_iNumChildren] = matChild;

    m_iNumChildren++;
}


int Model::GetNumChildren()
{
    return m_iNumChildren;
}

const D3DXMATRIX&Model::GetChildMatrix(int iChild)
{
    return m_vmatChild[iChild];
}

Model* Model::GetChild(int iChild)
{
    return m_vpChild[iChild];
}
