// Object.cpp: Implementierung der Klasse Object.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Object.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Object::~Object()
{
    for (unsigned i=0; i<m_vpChild.size();i++)
    {
        delete m_vpChild[i];
    }
}

Object::Object(Model *pModel)
{
    m_pModel = pModel;

    D3DXMatrixIdentity(&m_matWorld);

    m_vpChild = std::vector<Object*>(pModel->GetNumChildren());
    m_vmatChild = std::vector<D3DXMATRIX>(pModel->GetNumChildren());

    for (int i=0; i<pModel->GetNumChildren(); i++)
    {
        m_vmatChild[i] = pModel->GetChildMatrix(i);
        m_vpChild[i] = new Object(pModel->GetChild(i));
    }
}

void Object::Render()
{
    D3DXMATRIX matTmp;
    Render(*D3DXMatrixIdentity(&matTmp));

}

void Object::Render(const D3DXMATRIX&matParent)
{
    g_pD3DRenderLayer->SetRenderState( D3DRS_LIGHTING, 1 );

    D3DXMatrixMultiply(&m_matComposite, &m_matWorld, &matParent);

    m_pModel->Render(m_matComposite);

    for (unsigned i=0; i<m_vpChild.size(); i++)
    {
        m_vpChild[i]->Render(m_vmatChild[i] * m_matComposite);
    }
}

void Object::SetWorldMatrix(const D3DXMATRIX&matWorld)
{
    m_matWorld = matWorld;
}

Object* Object::GetChild(int iChild)
{
    return m_vpChild[iChild];
}

const D3DXMATRIX& Object::GetWorldMatrix()
{
    return m_matWorld;
}

const D3DMATRIX& Object::GetCompositeMatrix()
{
    return m_matComposite;
}
