// Parameters.cpp: Implementierung der Klasse Parameters.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Parameters.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Parameters::Parameters()
{
	m_pCurFile = NULL;
}

Parameters::~Parameters()
{

}

/**
 * Returns the file with _name if it exists, NULL otherwise
 */
File* Parameters::GetFile(std::string _name)
{
	File * f = NULL;

	for (unsigned i=0; i<m_vFile.size(); i++)
	{
		if (m_vFile[i].name == _name) f= &m_vFile[i];
	}

	// if file not yet loaded
	if (!f) 
	{
		// if file is successfully loaded, put it in f, otherwise f stays NULL
		if (LoadFile(_name)) f = &m_vFile[m_vFile.size()-1];
	}

	return f;
}

/**
 * Eliminates leading & trailing whitespaces
 */
void Parameters::Trim(std::string * _string)
{
	while (isspace(_string->c_str()[0]))
		*_string = _string->substr(1, _string->length()-1);


    while (isspace(_string->c_str()[_string->size()-1]))
        *_string = _string->substr(0, _string->length()-1);
}

int Parameters::LoadParameterInt(std::string _name, int _default)
{
	if (!m_pCurFile) return _default;

	const char * value = GetParameter(m_pCurFile, _name);

	if (value) return atoi(value);
	else return _default;
}

float Parameters::LoadParameterFloat(std::string _name, float _default)
{
	if (!m_pCurFile) return _default;

	const char * value = GetParameter(m_pCurFile, _name);

	if (value) return (float)atof(value);
	else return _default;
}

std::string Parameters::LoadParameterString(std::string _name, std::string _default)
{
	if (!m_pCurFile) return _default;

	const char * value = GetParameter(m_pCurFile, _name);

	if (value) return std::string(value);
	else return _default;
}

bool Parameters::LoadFile(std::string _name)
{
	ifstream in;
	File f;

	in.open(g_pUtils->GetFullPath((_name + ".txt")).c_str());

	if (in.fail())
	{
		std::string msg = "\tERROR: Failed to open parameterfile ";
		msg += _name;
		msg += ".";
		g_pUtils->Log(msg);
		return false;
	}

    std::string line;
    std::string name;
    std::string value;

    while (!in.eof())
    {
        try
        {
            std::getline(in, line);
            if (line.c_str()[0] == '#' || line == "") continue;

			Trim(&line);

            name = line.substr(0, line.find(" "));

            value = line.substr(line.find(" "), line.length());

			Trim(&name);
			Trim(&value);
        }
        catch (...)
        {
            std::string msg = "\tERROR: Parse error in file ";
            msg += _name;
            msg += ". Using default paramters.";
            g_pUtils->Log(msg);
            return false;
        }

		f.mapParams[name] = value;
	}

	f.name = _name;

	m_vFile.push_back(f);


	return true;
}

const char * Parameters::GetParameter(File * _file, std::string _name)
{
	assert(_file);

	std::map<std::string, std::string>::iterator i = _file->mapParams.find(_name);

	// found the parameter
	if (i != _file->mapParams.end())
	{
		return i->second.c_str();
	} else
	{
        std::string msg = "\tERROR: Parameter ";
		msg += _name + " not found in file " + _file->name + ". Using default parameter.";
        g_pUtils->Log(msg);	

		return NULL;
	}
}

/**
 *	Sets the current file to the file with the specified name
 */
void Parameters::SetFile(std::string _name)
{
	m_pCurFile = GetFile(_name);
}
