// ParticleSystem.cpp: Implementierung der Klasse ParticleSystem.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ParticleSystem.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////


ParticleSystem::ParticleSystem()
{

	if (g_pD3DLayer->GetD3DCaps().MaxPointSize == 1.0f || g_pD3DLayer->GetD3DCaps().MaxPointSize == 0.0f)
	{
		m_bPointSpritesSupported = false;
		m_bRenderPointSprites = false;
		g_pUtils->Log("Hardware PointSprites not supported, using billboard particles\n");
	} else
	{
		m_bPointSpritesSupported = true;
		m_bRenderPointSprites = true;
		g_pUtils->Log("Hardware PointSprites supported\n");
	}



    
	m_hVBPointSprites = g_pD3DRenderLayer->CreateVertexBuffer(3*VBSIZE, sizeof(PARTICLEVERTEXPOINT),
												  D3DUSAGE_DYNAMIC | D3DUSAGE_POINTS | D3DUSAGE_WRITEONLY,
												  D3DFVF_PARTICLEVERTEXPOINT);


	m_hVBBillboards = g_pD3DRenderLayer->CreateVertexBuffer(3*VBSIZE, sizeof(PARTICLEVERTEXBILLBOARD),
												  D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY,
												  D3DFVF_PARTICLEVERTEXBILLBOARD);

	D3DXMATRIX mat;
    /******* Begin state Block ************/
    g_pD3DRenderLayer->BeginStateBlock();

    g_pD3DRenderLayer->SetRenderState( D3DRS_LIGHTING, FALSE );
    g_pD3DRenderLayer->SetRenderState( D3DRS_POINTSPRITEENABLE, true );
    
    g_pD3DRenderLayer->SetRenderState( D3DRS_POINTSCALEENABLE,  true );
    g_pD3DRenderLayer->SetRenderState( D3DRS_POINTSIZE,     FtoDW(100.0f) );
    g_pD3DRenderLayer->SetRenderState( D3DRS_POINTSIZE_MIN, FtoDW(0.0f) );
    g_pD3DRenderLayer->SetRenderState( D3DRS_POINTSCALE_A,  FtoDW(1.00f) );
    g_pD3DRenderLayer->SetRenderState( D3DRS_POINTSCALE_B,  FtoDW(0.00f) );
    g_pD3DRenderLayer->SetRenderState( D3DRS_POINTSCALE_C,  FtoDW(0.1f) );

    g_pD3DRenderLayer->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );
    g_pD3DRenderLayer->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
    g_pD3DRenderLayer->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
    g_pD3DRenderLayer->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );

	g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE );
	g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_DIFFUSE );
	g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_TEXTURE );

    g_pD3DRenderLayer->SetTransform(D3DTS_WORLD, D3DXMatrixIdentity(&mat));

    g_pD3DRenderLayer->SetVertexShader( D3DFVF_PARTICLEVERTEXPOINT );

    g_pD3DRenderLayer->EndStateBlock(&m_dwPointSpriteStateBlock);  

	// billboard state block
	g_pD3DRenderLayer->BeginStateBlock();

    g_pD3DRenderLayer->SetRenderState( D3DRS_LIGHTING, FALSE );

    g_pD3DRenderLayer->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );
    g_pD3DRenderLayer->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
    g_pD3DRenderLayer->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
    g_pD3DRenderLayer->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );

	g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE );
	g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_DIFFUSE );
	g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_TEXTURE );

	g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_ADDRESSU, D3DTADDRESS_CLAMP );
	g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_ADDRESSV, D3DTADDRESS_CLAMP );

    g_pD3DRenderLayer->SetTransform(D3DTS_WORLD, D3DXMatrixIdentity(&mat));

    g_pD3DRenderLayer->SetVertexShader( D3DFVF_PARTICLEVERTEXBILLBOARD );

    g_pD3DRenderLayer->EndStateBlock(&m_dwBillboardStateBlock);

    /******** Undo StateBlock ****/
    g_pD3DRenderLayer->BeginStateBlock();

    g_pD3DRenderLayer->SetRenderState( D3DRS_POINTSPRITEENABLE, false );
    
    g_pD3DRenderLayer->SetRenderState( D3DRS_ZWRITEENABLE, true );
    g_pD3DRenderLayer->SetRenderState( D3DRS_ALPHABLENDENABLE, false );

	g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_DISABLE );
	g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_ADDRESSU, D3DTADDRESS_WRAP );
	g_pD3DRenderLayer->SetTextureStageState(0, D3DTSS_ADDRESSV, D3DTADDRESS_WRAP );


    g_pD3DRenderLayer->EndStateBlock(&m_dwUndoStateBlock); 
    /********* End state Block ************/

    m_pParticleFree = new PARTICLE;

    // allocate free particles
    PARTICLE * pCurParticle = m_pParticleFree;
    for (int i=0; i<MAXNUMPARTICLES-1; i++)
    {
        pCurParticle->pNext = new PARTICLE;
        pCurParticle = pCurParticle->pNext;
    }
    pCurParticle->pNext = NULL;

    srand(time(NULL));

	g_pD3DRenderLayer->GetViewport(&m_Viewport);
}

ParticleSystem::~ParticleSystem()
{
    g_pD3DRenderLayer->DeleteStateBlock(m_dwPointSpriteStateBlock);
}

void ParticleSystem::FrameMove()
{
	EFFECT * curEffect;

	for (unsigned e=0; e<m_vEffect.size(); e++)
	{
		curEffect = &m_vEffect[e];

		PARTICLE **ppParticle, *pParticle;
		ppParticle = &curEffect->pFirstParticle;

		while (*ppParticle)
		{
			pParticle = *ppParticle;

			pParticle->fLife -= curEffect->fLifeDecr;

			// kill expired particles
			if (pParticle->fLife <= 0)
			{
				*ppParticle = pParticle->pNext;
				pParticle->pNext = m_pParticleFree;
				m_pParticleFree = pParticle;
				curEffect->iNumParticles--;
			} else
			{
				pParticle->vPos += pParticle->vDir;
				pParticle->vDir *= curEffect->fDampening;
				pParticle->vDir += curEffect->vForce;
				D3DXCOLOR c;
				D3DXColorLerp(&c, &curEffect->cEnd, &curEffect->cBegin, pParticle->fLife);
				pParticle->c = (DWORD)c;

				ppParticle = &pParticle->pNext;
			}
		}
	}
}

void ParticleSystem::RenderPointSprites()
{
	EFFECT * curEffect;
    PARTICLEVERTEXPOINT * pCurVertex;
	PARTICLE * pCurParticle;
	bool first = true;

	for(unsigned e=0; e<m_vEffect.size(); e++)
	{
		curEffect = &m_vEffect[e];
		if (!curEffect->pFirstParticle) continue;

		if (first)
		{
			first = false;
			g_pD3DRenderLayer->SetStreamSource( 0, m_hVBPointSprites, sizeof(PARTICLEVERTEXPOINT) );
			g_pD3DRenderLayer->ApplyStateBlock(m_dwPointSpriteStateBlock);
		}

		g_pD3DRenderLayer->SetTexture(0, curEffect->hTex);

		g_pD3DRenderLayer->LockVB(m_hVBPointSprites, 0, curEffect->iNumParticles * sizeof(PARTICLEVERTEXPOINT),
								 (BYTE **) &pCurVertex, D3DLOCK_DISCARD);

		pCurParticle = curEffect->pFirstParticle;
		while(pCurParticle)
		{
			memcpy(pCurVertex++, pCurParticle, sizeof(PARTICLEVERTEXPOINT));
			pCurParticle = pCurParticle->pNext;
		}

	    g_pD3DRenderLayer->UnlockVB(m_hVBPointSprites);

	    g_pD3DRenderLayer->DrawPrimitive( D3DPT_POINTLIST, 0, curEffect->iNumParticles);

	}
   
    g_pD3DRenderLayer->ApplyStateBlock(m_dwUndoStateBlock);
     
}

//DEL void ParticleSystem::RenderChunks()
//DEL {
//DEL 	/*
//DEL     if (!m_pParticle) return;
//DEL 
//DEL 
//DEL     pd3dDevice->SetStreamSource( 0, m_pVB, sizeof(PARTICLEVERTEX) );
//DEL     pd3dDevice->ApplyStateBlock(m_dwParticleStateBlock);
//DEL 
//DEL 
//DEL     PARTICLEVERTEX * pCurVertex;
//DEL     int iParticlesInBuffer = 0;
//DEL 
//DEL 
//DEL     m_iVBBase += m_iChunkSize;
//DEL     if (m_iVBBase >= VBSIZE) m_iVBBase = 0;
//DEL 
//DEL 	m_pVB->Lock(m_iVBBase * sizeof(PARTICLEVERTEX), m_iChunkSize * sizeof(PARTICLEVERTEX),
//DEL                (BYTE **)&pCurVertex, m_iVBBase ? D3DLOCK_NOOVERWRITE : D3DLOCK_DISCARD);
//DEL 
//DEL 
//DEL 
//DEL     PARTICLE * pCurParticle = m_pParticle;
//DEL     while (pCurParticle)
//DEL     {
//DEL         memcpy(pCurVertex++, pCurParticle, sizeof(PARTICLEVERTEX));
//DEL         if (++iParticlesInBuffer == m_iChunkSize)
//DEL         {
//DEL             m_pVB->Unlock();
//DEL             pd3dDevice->DrawPrimitive( D3DPT_POINTLIST, m_iVBBase, m_iChunkSize);
//DEL 
//DEL             // lock next chunk. If VB is too small, discard and lock next VB
//DEL             m_iVBBase += m_iChunkSize;
//DEL             if (m_iVBBase >=VBSIZE) m_iVBBase = 0;
//DEL 
//DEL 	        m_pVB->Lock(m_iVBBase * sizeof(PARTICLEVERTEX), m_iChunkSize * sizeof(PARTICLEVERTEX),
//DEL                        (BYTE **)&pCurVertex, m_iVBBase ? D3DLOCK_NOOVERWRITE : D3DLOCK_DISCARD);
//DEL 
//DEL             iParticlesInBuffer = 0;
//DEL         }
//DEL 
//DEL         pCurParticle = pCurParticle->pNext;
//DEL     }
//DEL 
//DEL     m_pVB->Unlock();
//DEL 
//DEL     // render remaining particles
//DEL     pd3dDevice->DrawPrimitive( D3DPT_POINTLIST, m_iVBBase, iParticlesInBuffer);
//DEL 
//DEL 
//DEL 
//DEL     pd3dDevice->ApplyStateBlock(m_dwUndoStateBlock);
//DEL 	*/
//DEL }






void ParticleSystem::CreateParticles(handleEffect _eff, D3DXVECTOR3 _vPos, D3DXVECTOR3 _vMaxVelocity, float _randomness, int _num)
{
	EFFECT * e = &m_vEffect[_eff];

	for (int i=0; i<_num; i++)
	{
		if (!m_pParticleFree) return;

		e->iNumParticles++;

		// insert first free particle into active particle list
		PARTICLE * pNewParticle = m_pParticleFree;
		m_pParticleFree = m_pParticleFree->pNext;
		pNewParticle->pNext = e->pFirstParticle;
		e->pFirstParticle = pNewParticle;

		static const float f = 1.0f / (float)RAND_MAX;

		// set initial values
		pNewParticle->vDir = _vMaxVelocity*(float)rand()*f;
		pNewParticle->vPos = _vPos;
		pNewParticle->fLife = 1.0f;

		// calc random disturbance
		D3DXVECTOR3 dist = D3DXVECTOR3((float)rand()-(RAND_MAX>>1), 
									   (float)rand()-(RAND_MAX>>1),
									   (float)rand()-(RAND_MAX>>1));
		D3DXVec3Normalize(&dist, &dist);
		pNewParticle->vDir += dist * _randomness * (float)rand() * f;

	}
}

handleEffect ParticleSystem::CreateEffect(const char * _texName, float _lifedecr, D3DXCOLOR _colorbegin, D3DXCOLOR _colorend, float _dampening, D3DXVECTOR3 _force)
{
	EFFECT e;
	e.cBegin = D3DXCOLOR(_colorbegin);
	e.cEnd = D3DXCOLOR(_colorend);
	e.fDampening = _dampening;
	e.fLifeDecr = _lifedecr;
	e.hTex = g_pD3DRenderLayer->CreateTexture(_texName);
	e.pFirstParticle = NULL;
	e.vForce = _force;
	e.iNumParticles = 0;

	m_vEffect.push_back(e);

	return m_vEffect.size()-1;
}

void ParticleSystem::RenderBillboards(const Camera & _camera)
{
	EFFECT * curEffect;
    PARTICLEVERTEXBILLBOARD * pCurVertex;
	PARTICLE * pCurParticle;
	bool first = true;

	for(unsigned e=0; e<m_vEffect.size(); e++)
	{
		curEffect = &m_vEffect[e];
		if (!curEffect->pFirstParticle) continue;

		D3DXMATRIX matTrans;
		if (first)
		{
			first = false;
			g_pD3DRenderLayer->SetStreamSource( 0, m_hVBBillboards, sizeof(PARTICLEVERTEXBILLBOARD) );
			g_pD3DRenderLayer->ApplyStateBlock(m_dwBillboardStateBlock);


			D3DXMatrixMultiply(&matTrans, &_camera.m_matView, &_camera.m_matProjection);
		}

		g_pD3DRenderLayer->SetTexture(0, curEffect->hTex);

		g_pD3DRenderLayer->LockVB(m_hVBBillboards, 0, 3* curEffect->iNumParticles * sizeof(PARTICLEVERTEXBILLBOARD),
								 (BYTE **) &pCurVertex, D3DLOCK_DISCARD);

		

		pCurParticle = curEffect->pFirstParticle;
		int numToDraw = curEffect->iNumParticles;
		
		while(pCurParticle)
		{
			// transform particle pos into projection space
			D3DXVec3Transform(&pCurVertex->pos, &pCurParticle->vPos, &matTrans);

			// clip particles behind viewpoint
			if (pCurVertex->pos.z <0) 
			{
				pCurParticle = pCurParticle->pNext;
				numToDraw--;
				continue;
			}

			float invW = 1.0f / pCurVertex->pos.w;

			// transform into screen space

			// calc size factor based on z distance
			float fac = 3.0f * m_Viewport.Height * sqrtf(1.0f/(1.0f+0.0005f*pCurVertex->pos.z*pCurVertex->pos.z));
			// max. particle size
			if (fac > 35.0f) fac = 35.0f;

			pCurVertex->pos.x *= (float)(m_Viewport.Width>>1) * invW;
			pCurVertex->pos.x += m_Viewport.Width >>1;

			pCurVertex->pos.y *= -(float)(m_Viewport.Height>>1) * invW;
			pCurVertex->pos.y += (m_Viewport.Height >>1) - 3.0f*fac;

			pCurVertex->pos.z *= invW;
			pCurVertex->pos.w = 1.0f;

			pCurVertex->c = pCurParticle->c;
			pCurVertex->u = 0.5f;
			pCurVertex->v = -1.0f;

			
			pCurVertex[1].pos = pCurVertex->pos;
			pCurVertex[1].pos.x += 2.0f*fac;
			pCurVertex[1].pos.y += 4.0f*fac;
			pCurVertex[1].u = 1.5f;
			pCurVertex[1].v = 1.0f;
			pCurVertex[1].c = pCurVertex->c;


			pCurVertex[2].pos = pCurVertex->pos;
			pCurVertex[2].pos.x -= 2.0f*fac;
			pCurVertex[2].pos.y += 4.0f*fac;
			pCurVertex[2].u = -0.5f;
			pCurVertex[2].v = 1.0f;
			pCurVertex[2].c = pCurVertex->c;

			pCurVertex += 3;
			pCurParticle = pCurParticle->pNext;
		}

	    g_pD3DRenderLayer->UnlockVB(m_hVBBillboards);

	    if (numToDraw) g_pD3DRenderLayer->DrawPrimitive( D3DPT_TRIANGLELIST, 0, numToDraw );

	}
   
    g_pD3DRenderLayer->ApplyStateBlock(m_dwUndoStateBlock);
     
}

void ParticleSystem::ToggleRenderMode()
{
	if (m_bPointSpritesSupported) m_bRenderPointSprites ^= 1;
}

void ParticleSystem::Render(const Camera & _camera)
{
	if (m_bRenderPointSprites) RenderPointSprites();
	else RenderBillboards(_camera);
}
