// ParticleSystem.h: Schnittstelle fr die Klasse ParticleSystem.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_PARTICLESYSTEM_H__7DDF7C60_DD9D_11D5_ADB1_0020E061A328__INCLUDED_)
#define AFX_PARTICLESYSTEM_H__7DDF7C60_DD9D_11D5_ADB1_0020E061A328__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


#include "d3drenderlayer.h"
#include "camera.h"

#define VBSIZE 3072
#define MAXNUMPARTICLES 1024

#define PARTICLEPARAMSFILE "ParticleParams.txt"

class Game;

#define D3DFVF_PARTICLEVERTEXPOINT (D3DFVF_XYZ|D3DFVF_DIFFUSE)
struct PARTICLEVERTEXPOINT
{
    D3DXVECTOR3 pos;
    DWORD c;
};

#define D3DFVF_PARTICLEVERTEXBILLBOARD (D3DFVF_XYZRHW|D3DFVF_DIFFUSE|D3DFVF_TEX1|D3DFVF_TEXCOORDSIZE2(0))
struct PARTICLEVERTEXBILLBOARD
{
    D3DXVECTOR4 pos;
    DWORD c;
	float u, v;
};

struct PARTICLE
{
    D3DXVECTOR3 vPos;  //Matches pos & c from PARTICLEVERTEX
    DWORD c;		   // 
    D3DXVECTOR3 vDir;
    float fLife;        // ranges from 1.0f to 0.0f, s for color interpolation

    PARTICLE * pNext;
};

struct EFFECT
{
	D3DXCOLOR cBegin, cEnd;
	PARTICLE * pFirstParticle;
	float fLifeDecr;
	float fDampening;
	D3DXVECTOR3 vForce;
	handleTex hTex;
	int iNumParticles;
};


typedef int handleEffect;

#define g_pParticleSystem ParticleSystem::GetSingleton()
class ParticleSystem : public Singleton<ParticleSystem>
{
public:
	void Render(const Camera & _camera);
	void ToggleRenderMode();
	handleEffect CreateEffect(const char * _texName, float _lifedecr, D3DXCOLOR _colorbegin, D3DXCOLOR _colorend, float _dampening, D3DXVECTOR3 _force);
	void CreateParticles(handleEffect _eff, D3DXVECTOR3 _vPos, D3DXVECTOR3 _vMaxVelocity, float _randomness, int _num = 1);
	void FrameMove();
	ParticleSystem();
	virtual ~ParticleSystem();

private:
	bool m_bPointSpritesSupported;
	void RenderPointSprites();
	void RenderBillboards(const Camera & _camera);

	bool m_bRenderPointSprites;
	D3DVIEWPORT8 m_Viewport;
	int m_iVBBase;
	int m_iChunkSize;

	std::vector<EFFECT> m_vEffect;
	PARTICLE* m_pParticleFree;
	DWORD m_dwPointSpriteStateBlock;
	DWORD m_dwBillboardStateBlock;
    DWORD m_dwUndoStateBlock;
	handleVB m_hVBBillboards;
	handleVB m_hVBPointSprites;
};

#endif // !defined(AFX_PARTICLESYSTEM_H__7DDF7C60_DD9D_11D5_ADB1_0020E061A328__INCLUDED_)
