// ShadowVolume.cpp: Implementierung der Klasse ShadowVolume.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ShadowVolume.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

ShadowVolume::ShadowVolume(bool _static, int _maxNumSilVertices)
{
	m_iMaxNumSilVertices = _maxNumSilVertices;

	DWORD usage = D3DUSAGE_WRITEONLY;
	if (!_static) usage |= D3DUSAGE_DYNAMIC;

	m_hVB = g_pD3DRenderLayer->CreateVertexBuffer(m_iMaxNumSilVertices*2 +1, sizeof(D3DXVECTOR3), 
		usage, D3DFVF_SVVERTEX);

	// Create IB
	m_hIB = g_pD3DRenderLayer->CreateIndexBuffer(m_iMaxNumSilVertices*2+2 +m_iMaxNumSilVertices+2, usage);
	WORD * pIndices = new WORD[m_iMaxNumSilVertices*2+2 +m_iMaxNumSilVertices+2];
	WORD * curInd = pIndices;
	for (int i=0; i<m_iMaxNumSilVertices; i++)
	{
		*curInd++ = i+m_iMaxNumSilVertices;
		*curInd++ = i;
	}
	*curInd++ = m_iMaxNumSilVertices;
	*curInd++ = 0;

// circle cap
*curInd++ = 2*m_iMaxNumSilVertices;
for (i=0; i<m_iMaxNumSilVertices; i++)
{
	*curInd++ = m_iMaxNumSilVertices+i;
}
*curInd++ = m_iMaxNumSilVertices;


	g_pD3DRenderLayer->FillIB(m_hIB, pIndices, m_iMaxNumSilVertices*2+2 +m_iMaxNumSilVertices+2);
}

ShadowVolume::~ShadowVolume()
{

}

void ShadowVolume::SetSil(const std::vector<D3DXVECTOR3>& _sil, const D3DXVECTOR3 & _light, float length)
{
	m_iNumSilVertices = _sil.size();
	assert(m_iNumSilVertices <= m_iMaxNumSilVertices);

	D3DXVECTOR3 offset = _light * length;
	D3DXVECTOR3 * pProjSil = new D3DXVECTOR3[m_iNumSilVertices];
	for(int i=0; i<m_iNumSilVertices; i++)
		pProjSil[i] = offset + _sil[i];


	g_pD3DRenderLayer->FillVB(m_hVB, (BYTE*)(&_sil[0]), m_iNumSilVertices, sizeof(D3DXVECTOR3));
	g_pD3DRenderLayer->FillVB(m_hVB, (BYTE*)pProjSil, m_iNumSilVertices, sizeof(D3DXVECTOR3), m_iMaxNumSilVertices);

// cap volume
D3DXVECTOR3 center = (pProjSil[0]+pProjSil[_sil.size()/2]) /2;
g_pD3DRenderLayer->FillVB(m_hVB, (BYTE*)&center, 1, sizeof(D3DXVECTOR3), 2*m_iMaxNumSilVertices);

	delete [] pProjSil;
}

void ShadowVolume::Render()
{
	g_pD3DRenderLayer->SetIndices(m_hIB, 0);
	g_pD3DRenderLayer->SetStreamSource(0, m_hVB, sizeof(D3DXVECTOR3));
	g_pD3DRenderLayer->SetVertexShader(D3DFVF_SVVERTEX);
	g_pD3DRenderLayer->DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP, 0,
											m_iNumSilVertices+m_iMaxNumSilVertices,
											0, m_iNumSilVertices*2);

	// cap volume
	g_pD3DRenderLayer->DrawIndexedPrimitive(D3DPT_TRIANGLEFAN, m_iMaxNumSilVertices,
											m_iMaxNumSilVertices+1,
											m_iMaxNumSilVertices*2+2, m_iMaxNumSilVertices);
											
}
