// Sphere.cpp: Implementierung der Klasse Sphere.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Sphere.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

int Sphere::ms_iNumInstances = 0;
ShadowVolume * Sphere::ms_pSV = NULL;

Sphere::Sphere()
{
	if (!ms_iNumInstances)
	{
		SETPARAMETERFILE("GameParams");
		SETPARAMETERFILE(LOADSTRING("m_szBallParamsFile", "SphereParams"));

		m_fRadius     = LOADFLOAT("m_fRadius",           100.0f);

		m_fBounciness = LOADFLOAT("m_fBounciness",       0.9f);
		m_fDrag       = LOADFLOAT("m_fDrag",             0.999f);
		m_fRollDrag   = LOADFLOAT("m_fRollDrag",         0.98f);
		m_fRotateDrag = LOADFLOAT("m_fRotateDrag",       0.995f);

		m_fRollThreshold = LOADFLOAT("m_fRollThreshold", 5.0f);
		m_fStopThreshold = LOADFLOAT("m_fStopThreshold", 1.0f);


		std::string str;
		str = LOADSTRING("m_szBounceSound", "media/bounce.wav");
		m_hBounceSound = g_pDSoundLayer->CreateSound(str.c_str());

		str = LOADSTRING("m_szRollSound", "media/roll.wav");
		m_hRollSound = g_pDSoundLayer->CreateSound(str.c_str());


		D3DMATERIAL8 mtrl;
		ZeroMemory( &mtrl, sizeof(D3DMATERIAL8) );
    
		mtrl.Diffuse.r = mtrl.Ambient.r = 1.0f;
		mtrl.Diffuse.g = mtrl.Ambient.g = 1.0f;
		mtrl.Diffuse.b = mtrl.Ambient.b = 1.0f;
		mtrl.Diffuse.a = mtrl.Ambient.a = 1.0f;
        
		mtrl.Specular.r = 1.0f;
		mtrl.Specular.g = 1.0f;
		mtrl.Specular.b = 1.0f;
		mtrl.Specular.a = 1.0f;

		mtrl.Power = 30.0f;
    

		str = LOADSTRING("m_szSphereTexName", "media/ball.jpg");
		m_hTex = g_pD3DRenderLayer->CreateTexture(str.c_str());

		std::vector<MESHFACTORYVERTEX> vVertex;
		std::vector<WORD> vIndex;

		int tess = LOADINT("m_iSphereTesselation", 3);

		g_pMeshFactory->CreateGeosphere(tess,m_fRadius, 
										&vVertex, &vIndex);
		m_pMesh = g_pMeshManager->CreateMesh(vVertex, vIndex);

		
		CreateShadowVolume();


		str = LOADSTRING("m_szDustTexture", "media/dust.dds");

		m_hEffDust = g_pParticleSystem->CreateEffect(str.c_str(), 0.01f, 
													 0x40804000, 0x00804000, 0.99f,
													 D3DXVECTOR3(0.0f, -g_pGame->GetGravity()*0.2f, 0.0f));
	}

	ms_iNumInstances++;

	m_pVecOffset[0] = D3DXVECTOR3(0.0f, -m_fRadius, 0.0f);

	float fac = sqrtf(3.0f)/2;
	for (int i=1; i<9; i++)
	{
		m_pVecOffset[i]   = D3DXVECTOR3(cosf(i*D3DX_PI/4.0f)*m_fRadius*fac,
									    -m_fRadius/2.0f,
									    sinf(i*D3DX_PI/4.0f)*m_fRadius*fac);
		m_pVecOffset[i+8] = D3DXVECTOR3(cosf(i*D3DX_PI/4.0f)*m_fRadius,
									    0.0f,
									    sinf(i*D3DX_PI/4.0f)*m_fRadius);
	}

	m_bStopped = false;

	m_fRotAngle = 0.0f;

	m_vPos.x = LOADFLOAT("m_fStartX", 5000.0f);
	m_vPos.y = LOADFLOAT("m_fStartY", 5000.0f);
	m_vPos.z = LOADFLOAT("m_fStartZ", 3600.0f);

	m_vSpeed.x = LOADFLOAT("m_fSpeedX", 0.0f);
	m_vSpeed.y = LOADFLOAT("m_fSpeedY", 0.0f);
	m_vSpeed.z = LOADFLOAT("m_fSpeedZ", 0.0f);


	m_vRotAxis.x = LOADFLOAT("m_fRotAxisX", 1.0f);
	m_vRotAxis.y = LOADFLOAT("m_fRotAxisY", 0.0f);
	m_vRotAxis.z = LOADFLOAT("m_fRotAxisZ", 0.0f);

	m_fRotSpeed = LOADFLOAT("m_fRotSpeed", 0.0f);

	D3DXMatrixIdentity(&m_matRot);
	D3DXMatrixIdentity(&m_matPrevRot);
}

Sphere::~Sphere()
{	

	if (--ms_iNumInstances == 0)
	{
		SAFE_DELETE(ms_pSV);
	}
}

void Sphere::Render()
{
	g_pD3DRenderLayer->SetRenderState( D3DRS_WRAP0, D3DWRAPCOORD_0);
	g_pD3DRenderLayer->SetRenderState( D3DRS_SPECULARENABLE, 1 );

    g_pD3DRenderLayer->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
    g_pD3DRenderLayer->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
    g_pD3DRenderLayer->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_MODULATE );
	g_pD3DRenderLayer->SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE);
	g_pD3DRenderLayer->SetTexture(0, m_hTex);

	D3DXMATRIX matTrans, matComp, matRotComp;


	D3DXMatrixTranslation(&matTrans, m_vPos.x, m_vPos.y, m_vPos.z);
	D3DXMatrixRotationAxis(&m_matRot, &m_vRotAxis, -m_fRotAngle);

	D3DXMatrixMultiply(&matRotComp, &m_matPrevRot, &m_matRot);
	D3DXMatrixMultiply(&matComp, &matRotComp, &matTrans);

	g_pD3DRenderLayer->SetTransform(D3DTS_WORLD, &matComp);
	m_pMesh->Render();

	g_pD3DRenderLayer->SetRenderState( D3DRS_WRAP0, 0);
	g_pD3DRenderLayer->SetRenderState( D3DRS_SPECULARENABLE, 0 );

}



void Sphere::FrameMove()
{
	if (m_bStopped) return;

	static const D3DXVECTOR3 gravity = D3DXVECTOR3(0.0f, -g_pGame->GetGravity(), 0.0f);
	bool rolling=false;

	// handle reflection by terrain
	D3DXVECTOR3 testPoint;

	D3DXVECTOR3 n, right, forward;	// orthonormal vectors
	D3DXVECTOR3 w0;					// initial angular momentum
	float v0p, v0n;					// init. velocity parallel & normal to terrain
	float w0n, w0p;
	float v1pp, v1pn;

	D3DXVECTOR3 w1n, w1p, w1;

	for (int p=0; p<17; p++)
	{
		testPoint = m_vPos + m_pVecOffset[p];					// get one of the points on lower sphere
		if (g_pGame->GetTerrain()->Collide(&testPoint, &n))		// and check for collision (returns n)
		{
																// calc orthonormal vectors
			D3DXVec3Cross(&right, &n, &m_vSpeed);				
			D3DXVec3Normalize(&right, &right);					// right
			D3DXVec3Cross(&forward, &right, &n);				// forward
					


			v0p = D3DXVec3Dot(&forward, &m_vSpeed);				// horz. comp. of initial speed
			v0n = D3DXVec3Dot(&n    , &m_vSpeed);				// vert. comp. of initial speed


			// continue if ball is headed away
			if (v0n > 0.0f) continue;

			

			// check if ball is rolling
			if (-v0n < m_fRollThreshold)
			{
				rolling = true;
				m_vPos = testPoint - m_pVecOffset[p];	// move sphere out of terrain
				m_fRotSpeed *= m_fRollDrag;
				v0n = 0.0f;
			} else
			// calc new vertical speed
				v0n *= -m_fBounciness;

			// handle particles & sound & energy field
			float speed = D3DXVec3Length(&m_vSpeed);
			if (rolling)
			{
				if (n.y != 0.0f)
				{
					g_pParticleSystem->CreateParticles(m_hEffDust, m_vPos - n*m_fRadius, 0.0f*n, 2.0f);
					PlayBounceSound(speed, rolling);
				}
				else
				{
					D3DXVECTOR3 postrans;
					D3DXVec3TransformCoord(&postrans, &m_vPos, &g_pGame->GetCamera().m_matView);


					postrans /= speed*speed;
					g_pDSoundLayer->SetSoundPosition(postrans);
					g_pEnergy->Create(m_vPos - n*m_fRadius, n, D3DXVECTOR3(0.0f, 1.0f, 0.0f), speed*m_fRadius);
				}
			} else
			{
				// terrain
				if (n.y != 0.0f)
				{
					g_pParticleSystem->CreateParticles(m_hEffDust, m_vPos - n*m_fRadius, 0.2f*speed*n, 4.0f, 60);
					PlayBounceSound(speed, rolling);
				}
				// wall
				else
				{
					D3DXVECTOR3 postrans;
					D3DXVec3TransformCoord(&postrans, &m_vPos, &g_pGame->GetCamera().m_matView);


					postrans /= speed*speed;
					g_pDSoundLayer->SetSoundPosition(postrans);
					g_pEnergy->Create(m_vPos - n*m_fRadius, n, D3DXVECTOR3(0.0f, 1.0f, 0.0f), speed*m_fRadius);
				}

			}


			w0 = m_vRotAxis * m_fRotSpeed;						// initial ang. momentum

			w0p = D3DXVec3Dot(&forward, &w0);					// ang. mom. sideways to velocity
			w0n = D3DXVec3Dot(&right, &w0);						// ang. mom. in direction of velocity

			// handle ang.mom. vector in direction of velocity
			v1pp = 5.0f/7.0f * v0p - 2.0f/7.0f * w0n*m_fRadius;
			w1n = -v1pp / m_fRadius * right;
			
			// handle ang. mom. sideways to velocity
			v1pn = 2.0f/7.0f*w0p * m_fRadius;
			w1p =  v1pn/m_fRadius * forward;


			// sum up final velocity
			m_vSpeed = v1pp * forward + v1pn*right + v0n*n;

			// sum up final ang. mom.
			w1 = w1p + w1n + D3DXVec3Dot(&n, &w0)*n;

			m_fRotSpeed = D3DXVec3Length(&w1);
			D3DXVec3Normalize(&m_vRotAxis, &w1);

			m_fRotAngle = 0.0f;

			// combine current rotation and prev. rot. to new prev. rot.
			D3DXMatrixMultiply(&m_matPrevRot, &m_matPrevRot, &m_matRot);
			// set curr. rot to identity in case another point collides
			if (rolling) D3DXMatrixIdentity(&m_matRot);

			// check other points only if rolling
			if (!rolling) break;
		}
	}


	if (rolling)
		m_vSpeed *= m_fRollDrag;
	else
		m_fRotSpeed *= m_fRotateDrag;	

	m_vSpeed += gravity;
	m_vSpeed *= m_fDrag;
	m_vPos += m_vSpeed;
	m_fRotAngle += m_fRotSpeed;


	// stop sphere if speed is to small 
	static D3DXVECTOR3 oldPos = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
	oldPos = oldPos*0.95f + m_vPos*0.05f;
	D3DXVECTOR3 ab = oldPos - m_vPos;
	float delta = D3DXVec3Length(&ab);
	if (delta < m_fStopThreshold && rolling)
	{
		m_vSpeed *= 0.0f;
		m_fRotSpeed = 0.0f;
		m_bStopped = true;
	}

}


D3DXVECTOR3 Sphere::GetPos()
{
	return m_vPos;
}

void Sphere::CreateShadowVolume()
{
	int numShadowVertices = LOADINT("m_iNumShadowVertices", 20);

	ms_pSV = new ShadowVolume(false, numShadowVertices);

	D3DXVECTOR3 light, up=D3DXVECTOR3(0.0f, 1.0f, 0.0f), axis;

	light = g_pGame->GetLightDir();

	D3DXVec3Cross(&axis, &light, &up);
	float angle = asinf(sqrtf(light.x*light.x + light.z*light.z));

	D3DXMATRIX mat;
	D3DXMatrixRotationAxis(&mat, &axis, angle);

	std::vector<D3DXVECTOR3> vertex = std::vector<D3DXVECTOR3>(numShadowVertices);

	for (int i=0; i<numShadowVertices; i++)
	{
		float angle = (float)i*2.0f*D3DX_PI/(float)numShadowVertices;

		vertex[i].x = m_fRadius * cosf(angle);
		vertex[i].y = 0.0f;
		vertex[i].z = m_fRadius * sinf(angle);

		D3DXVec3TransformNormal(&vertex[i], &vertex[i], &mat);
	}

	ms_pSV->SetSil(vertex, light, 300*m_fRadius);
}

void Sphere::RenderShadow()
{
	D3DXMATRIX matTrans;
	D3DXMatrixTranslation(&matTrans, m_vPos.x, m_vPos.y, m_vPos.z);

	
	// render shadow stencil         
	g_pD3DRenderLayer->SetTransform(D3DTS_WORLDMATRIX(0), &matTrans);
	g_pD3DRenderLayer->SetRenderState(D3DRS_LIGHTING, false);       
	g_pD3DRenderLayer->SetRenderState(D3DRS_ZWRITEENABLE, false); 
	g_pD3DRenderLayer->SetRenderState(D3DRS_STENCILENABLE, true);       
    g_pD3DRenderLayer->SetRenderState( D3DRS_SHADEMODE,     D3DSHADE_FLAT );
    g_pD3DRenderLayer->SetRenderState( D3DRS_STENCILFUNC, D3DCMP_ALWAYS );

	// don't draw shadow volume
    g_pD3DRenderLayer->SetRenderState( D3DRS_ALPHABLENDENABLE, true );
    g_pD3DRenderLayer->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_ZERO );
    g_pD3DRenderLayer->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );

	g_pD3DRenderLayer->SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_INCR);

	ms_pSV->Render();

	g_pD3DRenderLayer->SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_DECR);
	g_pD3DRenderLayer->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);

	ms_pSV->Render();


    g_pD3DRenderLayer->SetRenderState( D3DRS_ALPHABLENDENABLE, false );
	g_pD3DRenderLayer->SetRenderState(D3DRS_LIGHTING, true );
	g_pD3DRenderLayer->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW );
	g_pD3DRenderLayer->SetRenderState(D3DRS_ZWRITEENABLE, true ); 
	g_pD3DRenderLayer->SetRenderState(D3DRS_STENCILENABLE, false );       
    g_pD3DRenderLayer->SetRenderState(D3DRS_SHADEMODE, D3DSHADE_GOURAUD );

}

D3DXVECTOR3 Sphere::GetVelocity()
{
	return m_vSpeed;
}

bool Sphere::HasStopped()
{
	return m_bStopped;
}

float Sphere::GetRadius()
{
	return m_fRadius;
}

void Sphere::PlayBounceSound(float speed, bool _rolling)
{
	D3DXVECTOR3 postrans;
	D3DXVec3TransformCoord(&postrans, &m_vPos, &g_pGame->GetCamera().m_matView);


	postrans /= speed*speed;
	g_pDSoundLayer->SetSoundPosition(postrans);
	g_pDSoundLayer->PlaySound(_rolling?m_hRollSound:m_hBounceSound);
}

void Sphere::SetVelocity(D3DXVECTOR3 _vel)
{
	m_vSpeed = _vel;
	m_bStopped = 0;
}

void Sphere::SetSoundPosition()
{
	D3DXVECTOR3 postrans;
	D3DXVec3TransformCoord(&postrans, &m_vPos, &g_pGame->GetCamera().m_matView);

	D3DXVec3Normalize(&postrans, &postrans);
	g_pDSoundLayer->SetSoundPosition(postrans);

}
