// Terrain.h: Schnittstelle fr die Klasse Terrain.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TERRAIN_H__7AFC2A00_988E_4F04_A69E_71E66D2BDA74__INCLUDED_)
#define AFX_TERRAIN_H__7AFC2A00_988E_4F04_A69E_71E66D2BDA74__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <fstream.h>
#include "globalsextern.h"
#include "d3drenderlayer.h"
#include "camera.h"
#include "math.h"
#include "game.h"

class Game;

#define D3DFVF_TERRAINVERTEX (D3DFVF_XYZ|D3DFVF_TEX2| \
							  D3DFVF_TEXCOORDSIZE2(0)|D3DFVF_TEXCOORDSIZE2(1))						  
struct TERRAINVERTEX
{
    D3DXVECTOR3 pos;
    float tu1, tv1;
    float tu2, tv2;
};

struct TerrainBlock
{
	int iBaseVertexIndex;
	int iMipMapLevel;
	float * pMaxDeltaY;
};

struct QuadNode
{
	AABB bb;
	D3DXVECTOR2 vGradient;
	QuadNode *pTL, *pTR, *pLL, *pLR;
	TerrainBlock * pBlock;
	unsigned char clipstat;
};

struct TerrainPatch
{
	QuadNode * pNodeTop;
	handleVB hVB;
	TerrainBlock * pTerrainBlock;	// pointer to beginning of TB-array
};


class Terrain  
{
public:
	D3DXVECTOR3 GetNormalAt(float _x, float _z, bool _normalize = true);
	float GetHeightAt(float _x, float _z);
	bool Collide(D3DXVECTOR3 *_pos, D3DXVECTOR3*_n = NULL);
	bool Collide(D3DXVECTOR3 const &_pos, float * _offset, D3DXVECTOR3 *_normal);
	void Render(Camera _camera);
	void LoadHeightmap(const char*_filename);
	Terrain();
	virtual ~Terrain();

private:
	D3DXVECTOR3 GetNormalAtCoords(float _x, float _z, bool _normalize = true);
	float GetHeightAtCoords(float _x, float _z);
	int m_iMaxPixelError;
	float m_fScaleWidth;
	float m_fScaleHeight;
	int m_iBlockSize;
	void DeleteNode(QuadNode * _node);
	void GetNeighbor(int _dir, TerrainPatch **_patch, TerrainBlock **_block);
	DWORD m_dwVertexShader;
	D3DXVECTOR3 m_vLightDir;
	BYTE * m_pHeightData;
	int m_iQuadTreeDepth;
	int m_iBlockRes;	// terrain resolution in mipmap blocks
	int m_iNumMipMapLevels;
	void CreateIndexBuffers();
	void CalcStaticLighting(D3DXVECTOR3 _lightDir, float _ambient);
    void BuildQuadTree(TerrainPatch * _patch, TERRAINVERTEX * _pGrid);
    int m_iPatchHeight;
	int m_iPatchWidth;
    int m_iTotalHeight;
	int m_iTotalWidth;
	int m_iNumPatchesX;
	int m_iNumPatchesZ;


	void CalcMaxDeltaY(TerrainBlock * _block, TERRAINVERTEX * _pTV);
    void RenderNode(TerrainPatch * _patch, QuadNode * _node, Camera _camera, bool _inside);
	void ProcessNode(TerrainPatch * _patch, QuadNode *_node, Camera _camera, bool _inside);
	void ProcessBlock(TerrainPatch * _patch, TerrainBlock* _block);
	void RenderBlock(TerrainPatch * _patch, TerrainBlock *_block);
	QuadNode * CreateQuadNode(TerrainPatch * _patch, int _depth, int _number, TERRAINVERTEX * _pTV, TERRAINVERTEX * _pGrid);
	handleIB * m_phIB;
	handleTex m_hGroundTex;
	handleTex m_hDetailTex;
	TerrainPatch * m_pPatch;

	int * m_pPrimitiveCount;	// number of primitives for each mipmaplevel
};

#endif // !defined(AFX_TERRAIN_H__7AFC2A00_988E_4F04_A69E_71E66D2BDA74__INCLUDED_)
