

#include "stdafx.h"
#include "globals.h"


void CleanUp()
{
    g_pUtils->Log("-----Cleaning up-----");

	SAFE_DELETE(g_pD3DFont);

	delete g_pParticleSystem;
	delete g_pEnergy;
    delete g_pGame;
    delete g_pD3DLayer;
    delete g_pD3DRenderLayer;
	delete g_pMeshManager;

    SetWindowPos(g_hWnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_HIDEWINDOW);

    g_pUtils->Log("-----Cleanup finished-----");
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    static POINT ptLastMouse;
    POINT ptMouse;
	static bool wire = 0;


	switch( message ) 
    {	
		case WM_MBUTTONDOWN:
			g_pGame->Move(1, 1);
			break;
        case WM_MOUSEMOVE:
            GetCursorPos(&ptMouse);		

            g_pGame->Pitch((float)(ptLastMouse.y - ptMouse.y)/1000.0f);
            g_pGame->Yaw((float)(ptMouse.x - ptLastMouse.x)/1000.0f);

            ptLastMouse = ptMouse;
            RECT rtWindow;
            GetWindowRect(hWnd, &rtWindow);

			// set cursor to center of screen if close to border
            if (ptMouse.x <= rtWindow.left+100 || ptMouse.y <= rtWindow.top+110 || 
                ptMouse.x >= rtWindow.right-100 || ptMouse.y >= rtWindow.bottom-110)
            {
                int avx = rtWindow.left + (rtWindow.right - rtWindow.left) /2;
                int avy = rtWindow.top + (rtWindow.bottom - rtWindow.top) /2;
                SetCursorPos(avx, avy);
                ptLastMouse.x = avx;
                ptLastMouse.y = avy;
            }

            break;

        case WM_KEYDOWN:
			switch(wParam)
			{
				case VK_CONTROL:
	                g_pGame->Move(1, -1);
					break;
				case VK_UP:
	                g_pGame->Move(2, 1);
					break;
				case VK_DOWN:
	                g_pGame->Move(2, -1);
					break;
				case VK_LEFT:
	                g_pGame->Move(0, -1);
					break;
				case VK_RIGHT:
	                g_pGame->Move(0, 1);
					break;
				case 80:	// 'p'
					g_pGame->TogglePause();
					break;
				case 83:	// 's'
					g_pD3DLayer->SaveScreenshot();
					break;
				case VK_ESCAPE:
					PostMessage(hWnd, WM_CLOSE, 0, 0);
					break;
				case 87:	// 'w'
					wire ^= 1;
					g_pD3DRenderLayer->SetRenderState( D3DRS_FILLMODE, wire?D3DFILL_WIREFRAME:D3DFILL_SOLID);
					break;
				case 66:	// 'b'
					g_pParticleSystem->ToggleRenderMode();
					break;
				case VK_SPACE:	
					g_pGame->Explode();
					break;
				case 70:	// 'f'
					g_pGame->ToggleFollow();
					break;
			}
            break;
        case WM_KEYUP:
			switch(wParam)
			{
				case VK_UP:
	                g_pGame->Move(2, 0);
					break;
				case VK_DOWN:
	                g_pGame->Move(2, 0);
					break;
				case VK_LEFT:
	                g_pGame->Move(0, 0);
					break;
				case VK_RIGHT:
	                g_pGame->Move(0, 0);
					break;
				case VK_CONTROL:
	                g_pGame->Move(1, 0);
					break;
			}
			break;
		case WM_MBUTTONUP:
			g_pGame->Move(1, 0);
			break;

        case WM_PAINT:
            ValidateRect(hWnd, NULL);
            break;
            
        case WM_SETCURSOR:
            if (FULLSCREEN) SetCursor(NULL); 
            else return DefWindowProc(hWnd, message, wParam, lParam);
            break;

		case WM_ACTIVATE:
			GetCursorPos(&ptLastMouse);
			break;

        case WM_CLOSE:
            CleanUp();
            break;

		default:
			return DefWindowProc( hWnd, message, wParam, lParam );
   }
   return 0;
}

void InitWindow( HINSTANCE hInstance )
{
    WNDCLASSEX wcex;

    wcex.cbSize = sizeof(WNDCLASSEX); 

    wcex.style			= CS_CLASSDC;
    wcex.lpfnWndProc	= (WNDPROC)WndProc;
    wcex.cbClsExtra		= 0;
    wcex.cbWndExtra		= 0;
    wcex.hInstance		= hInstance;
    wcex.hIcon			= LoadIcon(hInstance, (LPCTSTR)IDI_BOUNCE);
    wcex.hCursor		= NULL;
    wcex.hbrBackground	= NULL;
    wcex.lpszMenuName	= NULL;
    wcex.lpszClassName	= "Bounce";
    wcex.hIconSm		= LoadIcon(wcex.hInstance, (LPCTSTR)IDI_SMALL);

    RegisterClassEx(&wcex);

    HWND hWnd;

    g_hInstance = hInstance;

    g_pUtils->Log("Creating Window...");
        
    hWnd = CreateWindow("Bounce", "Bounce", WS_OVERLAPPEDWINDOW,
      0, 0, WINDOWWIDTH, WINDOWHEIGHT, GetDesktopWindow(), NULL, hInstance, NULL);
    g_hWnd = hWnd;

    if( !hWnd ) throw InitException("Unable to create window");

    ShowWindow( hWnd, SW_SHOWDEFAULT );
    UpdateWindow( hWnd );
}





int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow )
{
    // singletons
    Utils u;
	Parameters * p = new Parameters();

    g_pUtils->Log("-----Initialization-----");
    MeshFactory mf;
	DSoundLayer s;

	try
    {

	    InitWindow( hInstance ); 


		// layer singletons
        D3DLayer * d3dl = new D3DLayer(g_hWnd, g_hInstance);
        assert(d3dl);
        
        D3DRenderLayer * d3drl = new D3DRenderLayer(d3dl);
        assert(d3drl);


		// other singletons, Layers must have been created
		MeshManager * mm = new MeshManager();
		ParticleSystem *ps = new ParticleSystem();
		Energy * en = new Energy();

        // create D3DFONT
        HRESULT hres;	
        g_pD3DFont = new CD3DFont("Courier", 13);
        hres = g_pD3DFont->InitDeviceObjects(g_pD3DLayer->Getd3dDevice());
        if (FAILED(hres)) throw InitException(hres, "CD3DFont Initialization FAILED!");
        hres = g_pD3DFont->RestoreDeviceObjects();
        if (FAILED(hres)) throw InitException(hres, "CD3DFont Initialization FAILED!");


        Game * g = new Game();

        g_pUtils->Log("-----Initialization finished-----\n");
    } catch(InitException e)
    {
        g_pUtils->Log("-----Initialization FAILED-----");
        g_pUtils->Log("Error message: " + e.m_strErrorMsg + "\n");

        CleanUp();
    
        MessageBox(NULL, e.m_strErrorMsg.c_str(), "Initialization Error", MB_OK | MB_TASKMODAL);

        return -1;
    }

	// make sure parameterfiles aren't accessed after initialization
	delete p;

	LARGE_INTEGER li, li2;
	QueryPerformanceFrequency(&li);
	int minDur = (int)li.QuadPart / g_pGame->GetTargetFPS();
	QueryPerformanceCounter(&li2);

    MSG msg;
    try
    {
		// main loop
        while(1)
        {
            if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))       
                DispatchMessage(&msg);
        
			QueryPerformanceCounter(&li);

			if (li.QuadPart - li2.QuadPart >= minDur)
			{
				if (g_pGame)
				{
					g_pGame->FrameMove((int)(li.QuadPart - li2.QuadPart));
					g_pGame->Render();
				}
				else break;

				li2.QuadPart = li.QuadPart;
			}

        }
    }
    catch (...)
    {
        g_pUtils->Log("Unknown Exception caught in main loop");

        CleanUp();

        MessageBox(NULL, "Unknown Exception caught in main loop", "Error", MB_OK | MB_TASKMODAL);

        return -1;
    }

	return msg.wParam;
}